package de.fernunihagen.rtt.ui.internal;

import java.net.URL;

import java.util.HashMap;
import java.util.Map;
import java.util.Collections;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Plugin;
import org.eclipse.core.runtime.Status;

import org.eclipse.swt.graphics.Image;

import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.ITreeContentProvider;

import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;

import org.eclipse.ui.plugin.AbstractUIPlugin;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.Filter;
import org.osgi.framework.FrameworkUtil;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;

import org.osgi.util.tracker.ServiceTracker;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import de.fernunihagen.rtt.framework.RTTCore;

import de.fernunihagen.rtt.internal.PluginAdapter;

import de.fernunihagen.rtt.ui.IActionAdapterProvider;

/**
 * The activator class controls the plug-in life cycle
 */
@Contract
public final class RTTCoreUIPlugin extends AbstractUIPlugin {

	// The plug-in ID
	public static final String PLUGIN_ID = "de.fernunihagen.rtt.ui";

	// The shared instance
	private static RTTCoreUIPlugin plugin;

	private static final IPath _ICONS_PATH = new Path("/icons");
	private static final long SERVICE_TIMEOUT = 3000;

	private PluginAdapter<RTTCoreUIPlugin> _adapter;
	@Invar("null != $target")
	private Map<ImageDescriptor, Image> _imageCache;

	@Invar("null != $target")
	private String _activeTestResultContentProvider;
	@Invar("null != $target")
	private Filter _activeActionAdapterProvider;

	private ServiceTracker _actionAdapterProvider;

	/**
	 * The constructor
	 */
	public RTTCoreUIPlugin() {
		_adapter = null;
		_imageCache = Collections
				.synchronizedMap(new HashMap<ImageDescriptor, Image>());

		setActiveTestResultContentProvider("de.fernunihagen.rtt.ui.view.testresult.contentprovider");
		setActiveActionAdapterProvider("de.fernunihagen.rtt.ui.action.adapterprovider");
	}

	/**
	 * Sets the active action adapter provider.
	 * 
	 * @param name
	 *            Action adapter provider name
	 */
	@Pre("null != name")
	public synchronized void setActiveActionAdapterProvider(final String name) {
		_activeActionAdapterProvider = filterByComponentName(name);
	}

	@Post("null != $return")
	private synchronized Filter getActiveActionAdapterProvider() {
		return _activeActionAdapterProvider;
	}

	/**
	 * Sets the active test result content provider.
	 * 
	 * @param name
	 *            Test result content provider name
	 */
	@Pre("null != name")
	public synchronized void setActiveTestResultContentProvider(
			final String name) {
		_activeTestResultContentProvider = name;
	}

	@Post("null != $return")
	private synchronized String getActiveTestResultContentProvider() {
		return _activeTestResultContentProvider;
	}

	@Pre("null != name")
	@Post("null != $return")
	private Filter filterByComponentName(final String name) {
		try {
			return FrameworkUtil.createFilter(String.format("(%s=%s)",
					"component.name", name));
		} catch (InvalidSyntaxException e) {
			// unexpected
		}
		return null;
	}

	@Pre("null != tracker && null != filter")
	private ServiceReference findServiceReferenceByFilter(
			final ServiceTracker tracker, final Filter filter) {
		final ServiceReference[] references = tracker.getServiceReferences();
		for (ServiceReference reference : references) {
			if (filter.match(reference)) {
				return reference;
			}
		}

		getAdapter().logWarning(
				"No matching service reference, using first registered.");

		return tracker.getServiceReference();
	}

	@Pre("null != tracker && null != filter && null != clazz")
	private <T> T findServiceByFilter(final ServiceTracker tracker,
			final Filter filter, final Class<T> clazz) {
		try {
			tracker.waitForService(SERVICE_TIMEOUT);
		} catch (InterruptedException e) {
		}
		return clazz.cast(tracker.getService(findServiceReferenceByFilter(
				tracker, filter)));
	}

	@Override
	@Pre("$this.PLUGIN_ID == context.getBundle().getSymbolicName()")
	public void start(final BundleContext context) throws Exception {
		try {
			super.start(context);
			plugin = this;
			_adapter = new PluginAdapter<RTTCoreUIPlugin>(this,
					getRTTCorePlugin());
			_actionAdapterProvider = new ServiceTracker(context,
					IActionAdapterProvider.class.getName(), null);
			_actionAdapterProvider.open();

			getAdapter().logInfo(PLUGIN_ID + " started");
		} catch (Exception e) {
			getLog().log(new Status(IStatus.ERROR, PLUGIN_ID, null, e));
		}
	}

	@Override
	public void stop(final BundleContext context) throws Exception {
		_adapter.logInfo(PLUGIN_ID + " stopped");
		_adapter.close();
		_adapter = null;
		_actionAdapterProvider.close();

		synchronized (_imageCache) {
			for (Image image : _imageCache.values()) {
				image.dispose();
			}
			_imageCache.clear();
		}

		plugin = null;
		super.stop(context);
	}

	/**
	 * Retrieves the test result content provider.
	 * 
	 * @return Test result content provider
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If service unavailable</li>
	 *             <ul>
	 */
	@Post("null != $return")
	public ITreeContentProvider getTestResultContentProvider()
			throws CoreException {
		return (ITreeContentProvider) RTTCore.getTestRunListenerRegistry().get(
				getActiveTestResultContentProvider());
	}

	/**
	 * Retrieves the action adapter provider.
	 * 
	 * @return Action adapter provider
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If service unavailable</li>
	 *             <ul>
	 */
	@Post("null != $return")
	public IActionAdapterProvider getActionAdapterProvider()
			throws CoreException {
		IActionAdapterProvider provider = null;
		if (null != _actionAdapterProvider) {
			provider = findServiceByFilter(_actionAdapterProvider,
					getActiveActionAdapterProvider(),
					IActionAdapterProvider.class);
		}
		if (null == provider) {
			getAdapter().reportError(
					"Action adapter provider service unavailable");
		}
		return provider;
	}

	/**
	 * Returns the active workbench window
	 * 
	 * @return the active workbench window
	 */
	public IWorkbenchWindow getActiveWorkbenchWindow() {
		IWorkbench workBench = getDefault().getWorkbench();
		if (workBench == null) {
			return null;
		}
		return workBench.getActiveWorkbenchWindow();
	}

	/**
	 * Returns the active page in workbench window
	 * 
	 * @return the active page in workbench window
	 */
	public IWorkbenchPage getActivePage() {
		IWorkbenchWindow activeWorkbenchWindow = getActiveWorkbenchWindow();
		if (activeWorkbenchWindow == null) {
			return null;
		}
		return activeWorkbenchWindow.getActivePage();
	}

	/**
	 * Returns the shared instance
	 * 
	 * @return the shared instance
	 */
	public static RTTCoreUIPlugin getDefault() {
		return plugin;
	}

	/**
	 * Returns the adapted shared instance
	 * 
	 * @return the adapted shared instance
	 */
	public static PluginAdapter<? extends Plugin> getRTTCoreUIPlugin() {
		return getDefault().getAdapter();
	}

	@Pre("null != bundle && null != path")
	private ImageDescriptor createImageDescriptor(final Bundle bundle,
			final IPath path, final boolean useMissingImageDescriptor) {
		final URL url = FileLocator.find(bundle, path, null);
		if (url != null) {
			return ImageDescriptor.createFromURL(url);
		}
		if (useMissingImageDescriptor) {
			return ImageDescriptor.getMissingImageDescriptor();
		}
		return null;
	}

	@Pre("null != relativePath")
	@Post("null != $return")
	private ImageDescriptor _getImageDescriptor(final String relativePath) {
		IPath path = _ICONS_PATH.append(relativePath);
		return createImageDescriptor(getDefault().getBundle(), path, true);
	}

	@Pre("null != relativePath")
	@Post("null != $return")
	private Image _createImage(final String path) {
		final ImageDescriptor descriptor = _getImageDescriptor(path);
		Image image = _imageCache.get(descriptor);
		if (null == image) {
			image = descriptor.createImage();
			Image oldImage = _imageCache.put(descriptor, image);
			if (null != oldImage) {
				oldImage.dispose();
			}
		}
		return image;
	}

	/**
	 * Creates an image descriptor for the given path in a bundle. The path can
	 * contain variables like $NL$. If no image could be found, the 'missing
	 * image descriptor' is returned.
	 * 
	 * @param relativePath
	 *            Relative path to image
	 * 
	 * @return Image descriptor
	 */
	public static ImageDescriptor getImageDescriptor(final String relativePath) {
		return getDefault()._getImageDescriptor(relativePath);
	}

	/**
	 * Creates an image for the given path in a bundle.
	 * 
	 * @param path
	 *            Relative path to image
	 * 
	 * @return Image
	 */
	public static Image createImage(final String path) {
		return getDefault()._createImage(path);
	}

	private PluginAdapter<? extends Plugin> getAdapter() {
		if (null == _adapter) {
			return null;
		}		
		_adapter.init();
		return _adapter;
	}
}
