package de.fernunihagen.rtt.ui.internal.action;

import static de.fernunihagen.rtt.ui.internal.RTTCoreUIPlugin.getRTTCoreUIPlugin;
import static de.fernunihagen.rtt.ui.internal.RTTCoreUIPlugin.getImageDescriptor;

import java.lang.reflect.InvocationTargetException;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceDescription;
import org.eclipse.core.resources.IncrementalProjectBuilder;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.SubMonitor;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IAction;

import org.eclipse.jface.dialogs.ProgressMonitorDialog;

import org.eclipse.jface.operation.IRunnableWithProgress;

import org.eclipse.ltk.core.refactoring.Refactoring;
import org.eclipse.ltk.core.refactoring.RefactoringDescriptor;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;

import org.eclipse.ui.statushandlers.StatusManager;

import de.fernunihagen.rtt.framework.RTTCore;
import de.fernunihagen.rtt.framework.model.TestRootContext;

import de.fernunihagen.rtt.framework.repository.IRepository;
import de.fernunihagen.rtt.framework.testadapter.IRefactoringStrategy;
import de.fernunihagen.rtt.framework.testadapter.RefactoringCheckAll;
import de.fernunihagen.rtt.framework.testadapter.RefactoringTestElement;

/**
 * Refactoring action adapter.
 * 
 * @author Osama El Hosami
 */
@Contract
public class RefactoringActionAdapter extends AbstractActionAdapter {

	private static IStatus createError(final String message, final Throwable e) {
		return getRTTCoreUIPlugin().createError(message, e);
	}

	private void reportError(final String message, final Throwable e)
			throws CoreException {
		getRTTCoreUIPlugin().reportError(message, e);
	}

	private static void handleException(final CoreException e) {
		// Check for a nested CoreException
		IStatus status = e.getStatus();
		if (status != null && status.getException() instanceof CoreException) {
			status = ((CoreException) status.getException()).getStatus();
		}

		if (status == null) {
			StatusManager.getManager().handle(createError(e.getMessage(), e),
					StatusManager.SHOW);
		} else {
			StatusManager.getManager().handle(status, StatusManager.SHOW);
		}
	}

	private boolean canReRun(final Object target) {
		return null != target && target instanceof RefactoringTestElement
				&& null != ((RefactoringTestElement) target).getTarget();
	}

	private IWorkspaceDescription getWorkspaceDescription() {
		return getWorkspace().getDescription();
	}

	private void revertChanges(final IResource resource,
			final IProgressMonitor pm) throws CoreException {
		if (null == resource) {
			return;
		}

		try {
			final SubMonitor monitor = SubMonitor.convert(pm,
					"Reverting changes", 2);
			final IRepository repository = RTTCore.getRepositoryProvider()
					.getRepository(resource);
			final IResource[] resources = new IResource[] { resource };

			repository.revertChanges(resources, true, monitor.newChild(1));
			repository.refreshLocal(resources, IResource.DEPTH_INFINITE, monitor
					.newChild(1));
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Post("null != $return")
	private IWorkspace getWorkspace() {
		return getRTTCoreUIPlugin().getWorkspace();
	}

	private void build(final IProject project, final IProgressMonitor pm)
			throws CoreException {
		try {
			project.build(IncrementalProjectBuilder.INCREMENTAL_BUILD, pm);
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Override
	public IAction getTarget(final Object[] selected) {
		final RefactoringTestElement[] targetRef = { null };
		for (Object element : selected) {
			if (canReRun(element)) {
				targetRef[0] = (RefactoringTestElement) element;
				break;
			}
			if (element instanceof TestRootContext) {
				final TestRootContext context = (TestRootContext) element;
				final Object target = context.getTarget();
				if (canReRun(target)) {
					targetRef[0] = (RefactoringTestElement) target;
					break;
				}
			}
		}

		if (null != targetRef[0]) {
			final RefactoringTestElement element = targetRef[0];
			return new Action(getName(),
					getImageDescriptor("etool16/relaunch.gif")) {
				@Override
				public void run() {
					final ProgressMonitorDialog dialog = new ProgressMonitorDialog(
							null);
					final IRunnableWithProgress runnable = new IRunnableWithProgress() {
						public void run(final IProgressMonitor pm) {

							final SubMonitor monitor = SubMonitor.convert(pm,
									"Performing refactoring", 4);

							monitor.subTask(element.getLabel());

							try {
								final RefactoringDescriptor descriptor = element
										.getTarget().requestDescriptor(
												monitor.newChild(1));
								if (null == descriptor) {
									reportError(
											"Failed to (re-)create refactoring descriptor",
											null);
									return; // we should never reach this line
								}

								final String projectName = descriptor
										.getProject();
								final IResource projectResource = getWorkspace()
										.getRoot().findMember(projectName);

								revertChanges(projectResource, monitor
										.newChild(1));

								final RefactoringStatus status = new RefactoringStatus();
								final Refactoring refactoring = descriptor
										.createRefactoring(status);
								if (status.hasError()) {
									reportError(
											"Failed to (re-)create refactoring",
											null);
								}
								final IRefactoringStrategy strategy = new RefactoringCheckAll(
										refactoring);

								strategy.performChange(monitor.newChild(1));

								if (!getWorkspaceDescription().isAutoBuilding()) {
									final IProject project = projectResource
											.getProject();
									build(project, monitor.newChild(1));
								}

							} catch (CoreException e) {
								handleException(e);
							} finally {
								if (null != pm) {
									pm.done();
								}
							}
						}
					};

					try {
						dialog.run(true, true, runnable);
					} catch (InvocationTargetException e) {
						getRTTCoreUIPlugin().logError(
								e.getTargetException().getLocalizedMessage());
					} catch (InterruptedException e) {
						getRTTCoreUIPlugin().logInfo(
								"Operation canceled, through user request");
					}
				}
			};
		}
		return null;
	}

	@Override
	public boolean isApplicable(final Object[] selected) {
		for (Object element : selected) {
			if (canReRun(element)) {
				return true;
			}
			if (element instanceof TestRootContext) {
				final TestRootContext context = (TestRootContext) element;
				final Object target = context.getTarget();
				if (canReRun(target)) {
					return true;
				}
			}
		}
		return false;
	}
}
