package de.fernunihagen.rtt.ui.internal.view;

import static de.fernunihagen.rtt.ui.internal.RTTCoreUIPlugin.getRTTCoreUIPlugin;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Post;

import org.eclipse.core.runtime.Assert;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;

import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;

import de.fernunihagen.rtt.framework.ITestElement;
import de.fernunihagen.rtt.framework.ITestElementContainer;
import de.fernunihagen.rtt.framework.ITestRunListener;

import de.fernunihagen.rtt.framework.ITestElement.Result;

import de.fernunihagen.rtt.framework.model.TestRootContext;
import de.fernunihagen.rtt.framework.model.TestRootRunSession;
import de.fernunihagen.rtt.framework.model.TestRunEvent;

import de.fernunihagen.rtt.ui.internal.RTTCoreUIPlugin;

/**
 * Test result content provider.
 * 
 * @author Osama El Hosami
 */
public final class TestResultContentProvider implements ITreeContentProvider,
		ITestRunListener {

	private static final Object[] NO_ELEMENTS = new Object[0];

	private final static Color ROOT_FAILURE = new Color(RTTCoreUIPlugin
			.getDefault().getWorkbench().getDisplay(), 157, 172, 186);
	private final static Color ROOT_ERROR = new Color(RTTCoreUIPlugin
			.getDefault().getWorkbench().getDisplay(), 255, 128, 128);
	private final static Color ROOT_OK = new Color(RTTCoreUIPlugin.getDefault()
			.getWorkbench().getDisplay(), 129, 189, 114);

	private final static String PROP_ID = "component.name";
	private final static String PROP_NAME = "name";

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCoreUIPlugin().getLogger(
				TestResultContentProvider.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	private String _id;
	private String _name;
	private TreeViewer _viewer;

	/**
	 * Creates a test result content provider
	 */
	public TestResultContentProvider() {
		_viewer = null;
		_id = getClass().getName().toLowerCase();
		_name = getClass().getSimpleName().toLowerCase();
	}

	/**
	 * Component activation.
	 * 
	 * @param properties
	 *            Property map
	 */
	protected void activate(final Map<?, ?> properties) {
		_id = (String) properties.get(PROP_ID);

		if (properties.containsKey(PROP_NAME)) {
			_name = (String) properties.get(PROP_NAME);
		}

		if (getLogger().isDebugEnabled()) {
			final String message = String.format("[%s (%s) : activated",
					getID(), properties.toString());
			getLogger().debug(message);
		}
	}

	@Override
	public String getID() {
		return _id;
	}

	@Override
	public Object[] getChildren(final Object parentElement) {
		if (parentElement instanceof ITestElementContainer) {
			return ((ITestElementContainer) parentElement).getChildren();
		}
		return NO_ELEMENTS;
	}

	@Override
	public Object getParent(final Object element) {
		return ((ITestElement) element).getParentContainer();
	}

	@Override
	public boolean hasChildren(final Object element) {
		return (element instanceof ITestElementContainer)
				&& ((ITestElementContainer) element).hasChildren();
	}

	@Override
	public Object[] getElements(Object inputElement) {
		if (null != inputElement) {
			return getChildren(inputElement);
		}
		return NO_ELEMENTS;
	}

	@Override
	public void dispose() {
		_viewer = null;
	}

	@Override
	public synchronized void inputChanged(final Viewer viewer,
			final Object oldInput, final Object newInput) {
		_viewer = (TreeViewer) viewer;
	}

	private synchronized TreeViewer getViewer() {
		return _viewer;
	}

	@Post("null != $return")
	private Display getDisplay() {
		Display display;
		display = Display.getCurrent();
		if (display == null) {
			display = Display.getDefault();
		}
		return display;
	}

	private TestResultView showViewInActivePage() {
		final TestResultView[] viewPartRef = { null };
		getDisplay().syncExec(new Runnable() {
			@Override
			public void run() {
				viewPartRef[0] = TestResultView.showViewInActivePage();
			}
		});
		return viewPartRef[0];
	}

	private void setInput(final ITestElement root) {
		getDisplay().asyncExec(new Runnable() {
			@Override
			public void run() {
				final TreeViewer viewer = getViewer();
				viewer.setInput(root);
			}
		});
	}

	private void updateInfoField(final TestRootRunSession o,
			final TestResultView viewpart) {
		if (null == viewpart) {
			return;
		}

		int failures = 0;
		int errors = 0;
		int ignored = 0;

		for (ITestElement element : o.getChildren()) {
			Result result = element.getTestResult();
			if (result == Result.FAILURE) {
				failures++;
			} else if (result == Result.ERROR) {
				errors++;
			} else if (result == Result.IGNORED) {
				ignored++;
			}
		}

		// initialization should not count
		int testruns = o.getChildrenCount() - 1;
		if (testruns < 0) {
			testruns = 0;
		}

		final String formatStr = "Test-Runs: %d  Failures: %d  Errors: %d  Ignored: %d  Time: %.3fs  Progress-State: %s";
		final String text = String.format(formatStr, testruns, failures,
				errors, ignored, o.getElapsedTimeInSeconds(),
				o.getProgressState(false));

		final Label label = viewpart.getInfoField();

		if (0 != errors) {
			label.setBackground(ROOT_ERROR);
		} else if (0 != failures) {
			label.setBackground(ROOT_FAILURE);
		} else {
			label.setBackground(ROOT_OK);
		}

		label.setText(text);
		label.redraw();
	}

	@Override
	public void update(final TestRootRunSession o, final TestRunEvent arg) {
		final TestResultView viewpart = showViewInActivePage();

		if (null == getViewer()
				|| o.isDisposed()
				// Optimization to avoid unnecessary updates
				|| !(o instanceof TestRootRunSession || arg.getElement() instanceof TestRootContext)) {
			return;
		}

		final TreeViewer viewer = getViewer();
		Assert.isNotNull(viewer);

		if (null == viewer.getInput()) {
			setInput(o);
		}

		getDisplay().asyncExec(new Runnable() {
			@Override
			public void run() {
				switch (arg.getType()) {
				case FINISHED: {
					final List<ITestElement> updates = new ArrayList<ITestElement>(
							2);
					ITestElementContainer parent = arg.getElement()
							.getParentContainer();

					updates.add(arg.getElement());

					// walk up the chain to the top visible element
					ITestElementContainer container = parent;
					while (container != o) {
						parent = container;
						updates.add(parent);
						container = container.getParentContainer();
					}

					viewer.refresh(parent, false);
					viewer.update(updates.toArray(), null);
				}
					// fall through
				case STARTED:
				default:
					updateInfoField(o, viewpart);
					break;
				}
			}
		});
	}

	@Override
	public String getName() {
		return _name;
	}
}
