package de.fernunihagen.rtt.framework;

import org.eclipse.core.runtime.IAdaptable;

import de.fernunihagen.rtt.internal.filter.AndElementFilter;
import de.fernunihagen.rtt.internal.filter.CollectorFilterAdapter;
import de.fernunihagen.rtt.internal.filter.InvertedElementFilter;
import de.fernunihagen.rtt.internal.filter.OrElementFilter;
import de.fernunihagen.rtt.internal.filter.XorElementFilter;

import de.fernunihagen.rtt.internal.jdt.CollectorResourceFilterAdapter;

/**
 * Element filter operation class. Provides utility operations for filters.
 * 
 * @author Osama El Hosami
 */
public final class ElementFilterOp {

	private ElementFilterOp() {

	}

	/**
	 * Creates a resource filter matching elements in given collector element
	 * set.
	 * 
	 * @param <T>
	 *            Model element type
	 * 
	 * @param collector
	 *            Collector
	 * 
	 * @return Filter
	 */
	public static <T extends IAdaptable> IElementFilter<IAdaptable> adaptToResourceFilter(
			final IElementCollector<T> collector) {
		return new CollectorResourceFilterAdapter(collector);
	}

	/**
	 * Creates a filter matching elements in given collector element set.
	 * 
	 * @param <T>
	 *            Model element type
	 * 
	 * @param collector
	 *            Collector
	 * 
	 * @return Filter
	 */
	public static <T extends IAdaptable> IElementFilter<IAdaptable> adaptToElementFilter(
			final IElementCollector<T> collector) {
		return new CollectorFilterAdapter<IAdaptable>(collector);
	}

	/**
	 * Creates an inverted element filter.
	 * 
	 * @param <T>
	 *            Model element type
	 * 
	 * @param filter
	 *            to invert
	 * 
	 * @return Inverted element filter
	 */
	public static <T> IElementFilter<T> not(final IElementFilter<T> filter) {
		return new InvertedElementFilter<T>(filter);
	}

	/**
	 * Creates an and combined element filter.
	 * 
	 * @param <T>
	 *            Model element type
	 * 
	 * @param one
	 *            One filter
	 * @param other
	 *            Another filter
	 * 
	 * @return And combined element filter
	 */
	public static <T> IElementFilter<T> and(final IElementFilter<T> one,
			final IElementFilter<T> other) {
		return new AndElementFilter<T>(one, other);
	}

	/**
	 * Creates an or combined element filter.
	 * 
	 * @param <T>
	 *            Model element type
	 * 
	 * @param one
	 *            One filter
	 * @param other
	 *            Another filter
	 * 
	 * @return Or combined element filter
	 */
	public static <T> IElementFilter<T> or(final IElementFilter<T> one,
			final IElementFilter<T> other) {
		return new OrElementFilter<T>(one, other);
	}

	/**
	 * Creates an xor combined element filter.
	 * 
	 * @param <T>
	 *            Model element type
	 * 
	 * @param one
	 *            One filter
	 * @param other
	 *            Another filter
	 * 
	 * @return Xor combined element filter
	 */
	public static <T> IElementFilter<T> xor(final IElementFilter<T> one,
			final IElementFilter<T> other) {
		return new XorElementFilter<T>(one, other);
	}
}
