package de.fernunihagen.rtt.framework;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

/**
 * Test configuration interface.
 * 
 * @author Osama El Hosami
 */
@Contract
public interface ITestConfiguration {
	/**
	 * Configuration id.
	 * 
	 * @see #getID()
	 */
	String TAG_ID = ITestConfiguration.class.getName() + ".id";

	/**
	 * Creates a new child configuration of this configuration with the given
	 * type.
	 * 
	 * @param type
	 *            the type
	 * 
	 * @return a new child configuration
	 * 
	 * @see #getChild
	 * @see #getChildren
	 */
	@Pre("null != type")
	@Post("null != $return")
	ITestConfiguration createChild(String type);

	/**
	 * Creates a new child of this configuration with the given type and id. The
	 * id is stored in the child configuration.
	 * 
	 * @param type
	 *            the type
	 * @param id
	 *            the child id
	 * 
	 * @return a new child configuration with the given type and id
	 * 
	 * @see #getID
	 */
	@Pre("null != type && null != id")
	@Post("null != $return")
	ITestConfiguration createChild(String type, String id);

	/**
	 * Returns the first child with the given type id.
	 * 
	 * @param type
	 *            the type id
	 * 
	 * @return the first child with the given type
	 */
	@Pre("null != type")
	ITestConfiguration getChild(String type);

	/**
	 * Returns all children with the given type id.
	 * 
	 * @param type
	 *            the type id
	 * 
	 * @return an array of children with the given type
	 */
	@Pre("null != type")
	@Post("null != $return")
	ITestConfiguration[] getChildren(String type);

	/**
	 * Returns the floating point value of the given key.
	 * 
	 * @param key
	 *            the key
	 * @param defaultValue
	 *            default value
	 * 
	 * @return the value, or default value if the key was not found or was found
	 *         but was not a floating point number
	 */
	@Pre("null != key")
	Float get(String key, Float defaultValue);

	/**
	 * Returns the type for this configuration.
	 * 
	 * @return the configuration type
	 */
	@Post("null != $return")
	String getType();

	/**
	 * Returns the id for this configuration.
	 * 
	 * @return the configuration id, or <code>null</code> if none
	 * 
	 * @see #createChild(java.lang.String,java.lang.String)
	 */
	String getID();

	/**
	 * Returns the integer value of the given key.
	 * 
	 * @param key
	 *            the key
	 * @param defaultValue
	 *            default value
	 * 
	 * @return the value, or default value if the key was not found or was found
	 *         but was not an integer
	 */
	@Pre("null != key")
	Integer get(String key, Integer defaultValue);

	/**
	 * Returns the string value of the given key.
	 * 
	 * @param key
	 *            the key
	 * @param defaultValue
	 *            default value
	 * 
	 * @return the value, or defaultValue if the key was not found
	 */
	@Pre("null != key")
	String get(String key, String defaultValue);

	/**
	 * Returns the boolean value of the given key.
	 * 
	 * @param key
	 *            the key
	 * @param defaultValue
	 *            default value
	 * 
	 * @return the value, or default value if the key was not found
	 */
	@Pre("null != key")
	Boolean get(String key, Boolean defaultValue);

	/**
	 * Returns the data of the Text node of the configuration. Each
	 * configuration is allowed only one Text node.
	 * 
	 * @param defaultValue
	 *            default value
	 * 
	 * @return the data of the Text node of the configuration, or default value
	 *         if the configuration has no Text node.
	 */
	String get(String defaultValue);

	/**
	 * Returns an array of all the attribute keys of the configuration.
	 * 
	 * @return an array with all the attribute keys of the configuration
	 */
	@Post("null != $return")
	String[] getAttributeKeys();

	/**
	 * Sets the value of the given key to the given floating point number.
	 * 
	 * @param key
	 *            the key
	 * @param value
	 *            the value
	 */
	@Pre("null != key")
	void put(String key, float value);

	/**
	 * Sets the value of the given key to the given integer.
	 * 
	 * @param key
	 *            the key
	 * @param value
	 *            the value
	 */
	@Pre("null != key")
	void put(String key, int value);

	/**
	 * Copy the attributes and children from <code>configuration</code> to the
	 * receiver.
	 * 
	 * @param configuration
	 *            the configuration to be copied.
	 */
	@Pre("null != key && null != configuration")
	void put(ITestConfiguration configuration);

	/**
	 * Sets the value of the given key to the given string.
	 * 
	 * @param key
	 *            the key
	 * @param value
	 *            the value
	 */
	@Pre("null != key && null != value")
	void put(String key, String value);

	/**
	 * Sets the value of the given key to the given boolean value.
	 * 
	 * @param key
	 *            the key
	 * @param value
	 *            the value
	 */
	@Pre("null != key")
	void put(String key, boolean value);

	/**
	 * Sets the configuration's Text node to contain the given data. Creates the
	 * Text node if none exists. If a Text node does exist, it's current
	 * contents are replaced. Each configuration is allowed only one text node.
	 * 
	 * @param data
	 *            the data to be placed on the Text node
	 */
	@Pre("null != data")
	public void put(String data);

}
