package de.fernunihagen.rtt.framework;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import java.io.IOException;

import java.lang.annotation.Annotation;

import java.lang.reflect.Method;

import java.net.URL;

import java.security.AccessController;
import java.security.PrivilegedAction;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.Platform;

import org.osgi.framework.Bundle;

import de.fernunihagen.rtt.Test;
import de.fernunihagen.rtt.TestParam;

/**
 * Test description.
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class TestDescription extends AbstractTestDescription {
	public final static class ParamTypeInfo {
		public Annotation annotation;
		@Invar("null != $target")
		public final Class<?> type;

		@Pre("null != type")
		public ParamTypeInfo(final Class<?> type, final Annotation annotation) {
			this.annotation = annotation;
			this.type = type;
		}
	}

	private final String _symbolicName;
	private ClassLoader _classLoader;
	private final String[] _testCases;

	private static final class BundleClassLoader extends ClassLoader {
		@Invar("null != $target")
		private final Bundle _bundle;

		@Pre("null != bundle")
		public BundleClassLoader(final Bundle bundle) {
			_bundle = bundle;
		}

		@Override
		protected Class<?> findClass(final String name)
				throws ClassNotFoundException {
			return _bundle.loadClass(name);
		}

		@Override
		protected URL findResource(final String name) {
			return _bundle.getResource(name);
		}

		@SuppressWarnings("unchecked")
		@Override
		protected Enumeration<URL> findResources(final String name)
				throws IOException {
			return _bundle.getResources(name);
		}
	}

	private void reportError(final String message, final Throwable exception)
			throws CoreException {
		getRTTCorePlugin().reportError(message, exception);
	}

	@Pre("null != configuration")
	private String getSymbolicName(final ITestConfiguration configuration) {
		return configuration.get(KEY_NAME, (String) null);
	}

	@Post("null != $return")
	private Bundle getBundle() throws CoreException {
		final Bundle bundle = Platform.getBundle(_symbolicName);
		if (null == bundle) {
			reportError("Configuration error, plug-in not found", null);
		}
		return bundle;
	}

	@Pre("null != configuration")
	@Post("null != $return")
	private String[] getTestCases(final ITestConfiguration configuration) {
		final ITestConfiguration[] testcases = configuration
				.getChildren(TYPE_ACTORS_ACTOR_TESTBUNDLE_TESTCASE);
		final List<String> classNames = new ArrayList<String>(testcases.length);

		for (ITestConfiguration testcase : testcases) {
			final String className = testcase.get(KEY_NAME, (String) null);
			classNames.add(className);
		}

		return classNames.toArray(new String[classNames.size()]);
	}

	/**
	 * Creates a test run description.
	 * 
	 * @param configuration
	 *            Test configuration
	 */
	@Pre("null != configuration")
	public TestDescription(ITestConfiguration configuration) {
		super(configuration);

		// FIXME: extend handling for support of multiple bundles
		configuration = configuration.getChild(TYPE_ACTORS_ACTOR_TESTBUNDLE);
		_testCases = getTestCases(configuration);
		_symbolicName = getSymbolicName(configuration);
		_classLoader = null;
	}

	@Post("null != $return")
	private ClassLoader getClassLoader() throws CoreException {
		final Bundle bundle = getBundle();
		return AccessController
				.doPrivileged(new PrivilegedAction<BundleClassLoader>() {
					public BundleClassLoader run() {
						return new BundleClassLoader(bundle);
					}
				});
	}

	/**
	 * Loads a test class by name
	 * 
	 * @param name
	 *            Name of class
	 * 
	 * @return Class
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>On missconfiguration</li>
	 *             </ul>
	 */
	@Pre("null != name")
	@Post("null != $return")
	public Class<?> loadClass(final String name) throws CoreException {
		if (null == _classLoader) {
			_classLoader = getClassLoader();
		}

		Class<?> clazz = null;

		try {
			clazz = _classLoader.loadClass(name);
		} catch (ClassNotFoundException e) {
			reportError("Class not found", e);
		}

		return clazz;
	}

	/**
	 * Retrieves a list of test annotated methods in test cases.
	 * 
	 * @param testcases
	 *            List of test case classes
	 * 
	 * @return List of test methods
	 */
	public static Method[] getTests(final Class<?>[] testcases) {
		return getTests(testcases, Test.class);
	}

	/**
	 * Retrieves a list of annotated methods in test cases.
	 * 
	 * @param testcases
	 *            List of test case classes
	 * @param annotation
	 *            Annotation class
	 * 
	 * @return List of test methods
	 */
	public static Method[] getTests(final Class<?>[] testcases,
			final Class<? extends Annotation> annotation) {
		final List<Method> methods = new ArrayList<Method>();
		for (Class<?> testcase : testcases) {
			for (Method method : testcase.getMethods()) {
				if (method.isAnnotationPresent(annotation)) {
					methods.add(method);
				}
			}
		}
		return methods.toArray(new Method[methods.size()]);
	}

	/**
	 * Retrieves a list of test annotated parameters as parameter type
	 * information.
	 * 
	 * @param method
	 *            Method
	 * 
	 * @return List of parameter type information
	 */
	public static ParamTypeInfo[] getTestParams(final Method method) {
		return getTestParams(method, TestParam.class);
	}

	/**
	 * Retrieves a list of annotated parameters as parameter type information.
	 * 
	 * @param method
	 *            Method
	 * @param annotationClass
	 *            Annotation class
	 * 
	 * @return List of parameter type information
	 */
	public static ParamTypeInfo[] getTestParams(final Method method,
			final Class<? extends Annotation> annotationClass) {
		final Annotation[][] paramsAnnotations = method
				.getParameterAnnotations();
		final Class<?> paramTypes[] = method.getParameterTypes();
		final ParamTypeInfo[] params = new ParamTypeInfo[paramTypes.length];
		for (int i = 0; i < paramsAnnotations.length; i++) {
			Annotation annotation = null;
			for (Annotation paramAnnotation : paramsAnnotations[i]) {
				if (annotationClass.isInstance(paramAnnotation)) {
					annotation = paramAnnotation;
					break;
				}
			}
			params[i] = new ParamTypeInfo(paramTypes[i], annotation);
		}

		return params;
	}

	/**
	 * Retrieves the names of test case classes
	 * 
	 * @return Name of test case classes
	 */
	@Post("null != $return")
	public String[] getTestCases() {
		return _testCases;
	}
}
