package de.fernunihagen.rtt.framework;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.Writer;

import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * XML test configuration. Implementation based on
 * {@link org.eclipse.ui.XMLMemento}.
 * 
 * @author Osama El Hosami
 */
@Contract
public class XMLTestConfiguration implements ITestConfiguration {

	@Invar("null != $target")
	private final Document _factory;
	@Invar("null != $target")
	private final Element _element;

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				XMLTestConfiguration.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	private static void reportError(final String message,
			final Throwable exception) throws CoreException {
		getRTTCorePlugin().reportError(message, exception);
	}

	/**
	 * Creates a <code>Document</code> from the <code>Reader</code> and returns
	 * a configuration on the first <code>Element</code> for reading the
	 * document.
	 * <p>
	 * Same as calling createReadRoot(reader, null)
	 * </p>
	 * 
	 * @param reader
	 *            the <code>Reader</code> used to create the configuration
	 *            document
	 * 
	 * @return a configuration on the first <code>Element</code> for reading the
	 *         document
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>if IO problems</li>
	 *             <li>invalid format</li>
	 *             <li>or no element</li>
	 *             </ul>
	 */
	public static XMLTestConfiguration createReadRoot(final Reader reader)
			throws CoreException {
		return createReadRoot(reader, null);
	}

	/**
	 * Creates a <code>Document</code> from the <code>Reader</code> and returns
	 * a configuration on the first <code>Element</code> for reading the
	 * document.
	 * 
	 * @param reader
	 *            the <code>Reader</code> used to create the configuration's
	 *            document
	 * @param baseDir
	 *            the directory used to resolve relative file names in the XML
	 *            document. This directory must exist and include the trailing
	 *            separator. The directory format, including the separators,
	 *            must be valid for the platform. Can be <code>null</code> if
	 *            not needed.
	 * 
	 * @return a configuration on the first <code>Element</code> for reading the
	 *         document
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>if IO problems</li>
	 *             <li>invalid format</li>
	 *             <li>or no element</li>
	 *             </ul>
	 */
	public static XMLTestConfiguration createReadRoot(final Reader reader,
			final String baseDir) throws CoreException {
		try {
			final DocumentBuilderFactory factory = DocumentBuilderFactory
					.newInstance();
			final DocumentBuilder parser = factory.newDocumentBuilder();
			final InputSource source = new InputSource(reader);
			if (baseDir != null) {
				source.setSystemId(baseDir);
			}
			final Document document = parser.parse(source);
			final NodeList list = document.getChildNodes();
			for (int i = 0; i < list.getLength(); i++) {
				Node node = list.item(i);
				if (node instanceof Element) {
					return new XMLTestConfiguration(document, (Element) node);
				}
			}
		} catch (ParserConfigurationException e) {
			reportError("Internal XML parser configuration error", e);
		} catch (IOException e) {
			reportError("Could not read content of XML file", e);
		} catch (SAXException e) {
			reportError("Could not parse content of XML file", e);
		}
		reportError("Could not find root element node of XML file", null);
		return null;
	}

	/**
	 * Returns a root configuration for writing a document.
	 * 
	 * @param type
	 *            the element node type to create on the document
	 * 
	 * @return the root configuration for writing a document
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	public static XMLTestConfiguration createWriteRoot(final String type)
			throws CoreException {
		final Document document;
		try {
			document = DocumentBuilderFactory.newInstance()
					.newDocumentBuilder().newDocument();
			Element element = document.createElement(type);
			document.appendChild(element);
			return new XMLTestConfiguration(document, element);
		} catch (ParserConfigurationException e) {
			reportError(null, e);
		}
		return null;
	}

	/**
	 * Creates a configuration for the specified document and element.
	 * <p>
	 * Clients should use <code>createReadRoot</code> and
	 * <code>createWriteRoot</code> to create the initial configuration on a
	 * document.
	 * </p>
	 * 
	 * @param document
	 *            the document for the configuration
	 * @param element
	 *            the element node for the configuration
	 */
	@Pre("null != document && null != element")
	public XMLTestConfiguration(final Document document, final Element element) {
		_factory = document;
		_element = element;
	}

	@Override
	@Pre
	@Post
	public XMLTestConfiguration createChild(final String type) {
		final Element child = _factory.createElement(type);
		_element.appendChild(child);
		return new XMLTestConfiguration(_factory, child);
	}

	@Override
	@Pre
	@Post
	public XMLTestConfiguration createChild(final String type, final String id) {
		final Element child = _factory.createElement(type);
		child.setAttribute(TAG_ID, id == null ? "" : id);
		_element.appendChild(child);
		return new XMLTestConfiguration(_factory, child);
	}

	@Pre("null != child")
	@Post("null != $return")
	public XMLTestConfiguration copyChild(final XMLTestConfiguration child) {
		final Element childElement = child._element;
		final Element newElement = (Element) _factory.importNode(childElement,
				true);
		_element.appendChild(newElement);
		return new XMLTestConfiguration(_factory, newElement);
	}

	/**
	 * Removes this configuration node from configuration.
	 */
	public void remove() {
		_element.getParentNode().removeChild(_element);
	}

	/**
	 * Removes the first child with given type.
	 * 
	 * @param type
	 *            Type
	 * 
	 * @return True if succeeded otherwise false
	 */
	public boolean removeChild(final String type) {
		final Element element = _getChild(type);
		if (null == element) {
			return false;
		}

		element.getParentNode().removeChild(element);
		return true;
	}

	private Element _getChild(final String type) {
		// Get the nodes.
		final NodeList nodes = _element.getChildNodes();
		int size = nodes.getLength();
		if (size == 0) {
			return null;
		}

		// Find the first node which is a child of this node.
		for (int nX = 0; nX < size; nX++) {
			final Node node = nodes.item(nX);
			if (node instanceof Element) {
				final Element element = (Element) node;
				if (element.getNodeName().equals(type)) {
					return element;
				}
			}
		}

		// A child was not found.
		return null;
	}

	@Override
	@Pre
	public XMLTestConfiguration getChild(final String type) {
		final Element element = _getChild(type);
		if (null == element) {
			return null;
		}

		return new XMLTestConfiguration(_factory, element);
	}

	/**
	 * Removes children of given type from configuration.
	 * 
	 * @param type
	 *            Type
	 */
	public void removeChildren(final String type) {
		final Element[] elements = _getChildren(type);
		for (Element element : elements) {
			element.getParentNode().removeChild(element);
		}
	}

	private Element[] _getChildren(final String type) {
		// Get the nodes.
		final NodeList nodes = _element.getChildNodes();
		int size = nodes.getLength();
		if (size == 0) {
			return new Element[0];
		}

		// Extract each node with given type.
		final List<Element> list = new ArrayList<Element>(size);
		for (int nX = 0; nX < size; nX++) {
			final Node node = nodes.item(nX);
			if (node instanceof Element) {
				final Element element = (Element) node;
				if (element.getNodeName().equals(type)) {
					list.add(element);
				}
			}
		}

		return list.toArray(new Element[list.size()]);
	}

	@Override
	@Pre
	@Post
	public XMLTestConfiguration[] getChildren(final String type) {
		final Element[] elements = _getChildren(type);
		final XMLTestConfiguration[] results = new XMLTestConfiguration[elements.length];
		for (int x = 0; x < elements.length; x++) {
			results[x] = new XMLTestConfiguration(_factory, elements[x]);
		}
		return results;
	}

	@Override
	@Pre
	public Float get(final String key, final Float defaultValue) {
		final Attr attr = _element.getAttributeNode(key);
		if (attr != null) {
			final String strValue = attr.getValue();
			try {
				return new Float(strValue);
			} catch (NumberFormatException e) {
				getLogger().debug(
						"Configuration problem - Invalid float for key '" + key
								+ "' value '" + strValue + "'", e);
			}
		}
		return defaultValue;
	}

	@Override
	@Post
	public String getType() {
		return _element.getNodeName();
	}

	@Override
	public String getID() {
		return _element.getAttribute(TAG_ID);
	}

	@Override
	@Pre
	public Integer get(final String key, final Integer defaultValue) {
		final Attr attr = _element.getAttributeNode(key);
		if (attr != null) {
			final String strValue = attr.getValue();
			try {
				return new Integer(strValue);
			} catch (NumberFormatException e) {
				getLogger().debug(
						"Configuration problem - Invalid integer for key '"
								+ key + "' value '" + strValue + "'", e);
			}
		}
		return defaultValue;
	}

	@Override
	@Pre
	public String get(final String key, final String defaultValue) {
		final Attr attr = _element.getAttributeNode(key);
		if (attr == null) {
			return defaultValue;
		}
		return attr.getValue();
	}

	@Override
	@Pre
	public Boolean get(final String key, final Boolean defaultValue) {
		final Attr attr = _element.getAttributeNode(key);
		if (attr == null) {
			return defaultValue;
		}
		return Boolean.valueOf(attr.getValue());
	}

	@Override
	public String get(final String defaultValue) {
		final Text textNode = getTextNode();
		if (textNode != null) {
			return textNode.getData();
		}
		return defaultValue;
	}

	@Override
	@Post
	public String[] getAttributeKeys() {
		final NamedNodeMap map = _element.getAttributes();
		final int size = map.getLength();
		final String[] attributes = new String[size];
		for (int i = 0; i < size; i++) {
			final Node node = map.item(i);
			attributes[i] = node.getNodeName();
		}
		return attributes;
	}

	/**
	 * Returns the Text node of the configuration. Each configuration is allowed
	 * only one Text node.
	 * 
	 * @return the Text node of the configuration, or <code>null</code> if the
	 *         configuration has no Text node.
	 */
	private Text getTextNode() {
		// Get the nodes.
		final NodeList nodes = _element.getChildNodes();
		final int size = nodes.getLength();
		if (size == 0) {
			return null;
		}
		for (int nX = 0; nX < size; nX++) {
			final Node node = nodes.item(nX);
			if (node instanceof Text) {
				return (Text) node;
			}
		}
		// a Text node was not found
		return null;
	}

	/**
	 * Places the element's attributes into the document.
	 * 
	 * @param copyText
	 *            true if the first text node should be copied
	 */
	private void putElement(final Element element, final boolean copyText) {
		final NamedNodeMap nodeMap = element.getAttributes();
		int size = nodeMap.getLength();
		for (int i = 0; i < size; i++) {
			final Attr attr = (Attr) nodeMap.item(i);
			put(attr.getName(), attr.getValue());
		}

		final NodeList nodes = element.getChildNodes();
		size = nodes.getLength();

		boolean needToCopyText = copyText;
		for (int i = 0; i < size; i++) {
			Node node = nodes.item(i);
			if (node instanceof Element) {
				final XMLTestConfiguration child = createChild(node
						.getNodeName());
				child.putElement((Element) node, true);
			} else if (node instanceof Text && needToCopyText) {
				put(((Text) node).getData());
				needToCopyText = false;
			}
		}
	}

	@Override
	@Pre
	public void put(final String key, final float f) {
		_element.setAttribute(key, String.valueOf(f));
	}

	@Override
	@Pre
	public void put(final String key, final int n) {
		_element.setAttribute(key, String.valueOf(n));
	}

	@Override
	@Pre
	public void put(final ITestConfiguration configuration) {
		// Do not copy the element's top level text node (this would overwrite
		// the existing text).
		// Text nodes of children are copied.
		putElement(((XMLTestConfiguration) configuration)._element, false);
	}

	@Override
	@Pre
	public void put(final String key, final String value) {
		if (value == null) {
			return;
		}
		_element.setAttribute(key, value);
	}

	@Override
	@Pre
	public void put(final String key, final boolean value) {
		_element.setAttribute(key, value ? "true" : "false");
	}

	@Override
	@Pre
	public void put(final String data) {
		Text textNode = getTextNode();
		if (textNode == null) {
			textNode = _factory.createTextNode(data);
			_element.insertBefore(textNode, _element.getFirstChild());
		} else {
			textNode.setData(data);
		}
	}

	/**
	 * Saves this configuration document current values to the specified writer.
	 * 
	 * @param writer
	 *            the writer used to save the configuration's document
	 * 
	 * @return true if operation successful
	 */
	public boolean save(final Writer writer) {
		final DOMWriter out = new DOMWriter(writer);
		out.print(_element);
		final boolean fError = out.checkError();
		out.close();
		return !fError;
	}

	/**
	 * A simple XML writer. Taken from XMLMemento implementation.
	 */
	private static final class DOMWriter extends PrintWriter {
		private int _tab;

		private static final String XML_VERSION = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"; //$NON-NLS-1$

		public DOMWriter(final Writer output) {
			super(output);
			_tab = 0;
			print(XML_VERSION);
		}

		public void print(final Element element) {
			// Ensure extra whitespace is not emitted next to a Text node,
			// as that will result in a situation where the restored text data
			// is not the
			// same as the saved text data.
			boolean hasChildren = element.hasChildNodes();
			startTag(element, hasChildren);
			if (hasChildren) {
				_tab++;
				NodeList children = element.getChildNodes();
				for (int i = 0; i < children.getLength(); i++) {
					Node node = children.item(i);
					if (node instanceof Element) {
						print((Element) children.item(i));
					} else if (node instanceof Text) {
						print(getEscaped(node.getNodeValue().trim()));
					}
				}
				_tab--;
				endTag(element);
			}
		}

		private void startTag(final Element element, final boolean hasChildren) {
			final StringBuffer sb = new StringBuffer();
			sb.append("\n<");
			sb.append(element.getTagName());
			final NamedNodeMap attributes = element.getAttributes();
			for (int i = 0; i < attributes.getLength(); i++) {
				final Attr attribute = (Attr) attributes.item(i);
				sb.append(" ");
				sb.append(attribute.getName());
				sb.append("=\"");
				sb.append(getEscaped(String.valueOf(attribute.getValue())));
				sb.append("\"");
			}
			sb.append(hasChildren ? ">" : "/>");
			print(sb.toString());
		}

		private void endTag(final Element element) {
			final StringBuffer sb = new StringBuffer();
			sb.append("\n</");
			sb.append(element.getNodeName());
			sb.append(">");
			print(sb.toString());
		}

		private static void appendEscapedChar(final StringBuffer buffer,
				final char c) {
			final String replacement = getReplacement(c);
			if (replacement != null) {
				buffer.append('&');
				buffer.append(replacement);
				buffer.append(';');
			} else if (c == 9 || c == 10 || c == 13 || c >= 32) {
				buffer.append(c);
			}
		}

		private static String getEscaped(final String s) {
			final StringBuffer result = new StringBuffer(s.length() + 10);
			for (int i = 0; i < s.length(); ++i) {
				appendEscapedChar(result, s.charAt(i));
			}
			return result.toString();
		}

		private static String getReplacement(final char c) {
			switch (c) {
			case '':
				return "#228";
			case '':
				return "#246";
			case '':
				return "#252";
			case '':
				return "#196";
			case '':
				return "#214";
			case '':
				return "#220";
			case '<':
				return "lt";
			case '>':
				return "gt";
			case '"':
				return "quot";
			case '\'':
				return "apos";
			case '&':
				return "amp";
			case '\r':
				return "#x0D";
			case '\n':
				return "#x0A";
			case '\u0009':
				return "#x09";
			}
			return null;
		}
	}

}
