package de.fernunihagen.rtt.framework.model;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import de.fernunihagen.rtt.framework.ITestElement;
import de.fernunihagen.rtt.framework.ITestRunSession;

/**
 * Abstract test element.
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class AbstractTestElement implements ITestElement {

	@Invar("null != $target")
	private ProgressState _state;
	private long _elapsed;
	private long _startTime;

	/**
	 * Creates a test element.
	 */
	public AbstractTestElement() {
		_state = ProgressState.NOT_STARTED;
		_elapsed = 0;
		_startTime = 0;
	}

	@Override
	public Object getTarget() {
		return getParentContainer().getTarget();
	}

	@Override
	@Post
	public long getElapsedTimeInTicks() {
		return _elapsed;
	}

	@Override
	@Post
	public double getElapsedTimeInSeconds() {
		return Double.valueOf(getElapsedTimeInTicks()) / 1000.0f;
	}

	@Override
	@Post
	public ProgressState getProgressState(final boolean includeChilds) {
		return _state;
	}

	@Override
	@Post
	public RunStage getStage() {
		return getParentContainer().getStage();
	}

	/**
	 * Reset test element for new run.
	 * 
	 * @return true if operation succeeded otherwise false
	 */
	protected boolean reset() {
		_elapsed = 0;
		_startTime = 0;
		return true;
	}

	@Pre("null != state")
	private boolean notStarted(final ProgressState state) {
		boolean fTransition = true;

		switch (state) {
		case RUNNING:
			_startTime = System.currentTimeMillis();
			break;
		case COMPLETED:
			_elapsed = 0;
			break;
		default:
			fTransition = false;
			break;
		}

		return fTransition;
	}

	@Pre("null != state")
	private boolean stopped(final ProgressState state) {
		boolean fTransition = true;

		switch (state) {
		case RUNNING:
			_startTime = System.currentTimeMillis();
			break;
		case NOT_STARTED:
			fTransition = reset();
			break;
		default:
			fTransition = false;
			break;
		}

		return fTransition;
	}

	@Pre("null != state")
	private boolean running(final ProgressState state) {
		boolean fTransition = true;

		switch (state) {
		case STOPPED:
			// fall through
		case COMPLETED:
			_elapsed += System.currentTimeMillis() - _startTime;
			break;
		default:
			fTransition = false;
			break;
		}

		return fTransition;
	}

	@Pre("null != state")
	private boolean completed(final ProgressState state) {
		boolean fTransition = true;

		switch (state) {
		case RUNNING:
			_startTime = System.currentTimeMillis();
			break;
		case NOT_STARTED:
			fTransition = reset();
			break;
		default:
			fTransition = false;
			break;
		}

		return fTransition;
	}

	/**
	 * Sets the progress state of run context.
	 * 
	 * @param state
	 *            Progress state
	 */
	@Pre("null != state")
	protected void setProgressState(final ProgressState state) {
		boolean fTransition = true;
		final ProgressState actState = getProgressState(false);

		switch (actState) {
		case NOT_STARTED:
			fTransition = notStarted(state);
			break;
		case STOPPED:
			fTransition = stopped(state);
			break;
		case RUNNING:
			fTransition = running(state);
			break;
		case COMPLETED:
			fTransition = completed(state);
			break;
		default:
			fTransition = false;
			break;
		}

		if (fTransition) {
			_state = state;
			getSession().notifyChange(this, actState);
		}
	}

	@Override
	@Post
	public ITestRunSession getSession() {
		return getParentContainer().getSession();
	}

	@Override
	@Post("null != $return")
	public String toString() {
		final String message = String.format("[%s (/L %s)]",
				getParentContainer(), getLabel());
		return message;
	}
}
