package de.fernunihagen.rtt.framework.model;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import de.fernunihagen.rtt.framework.ITestElement;
import de.fernunihagen.rtt.framework.ITestElementContainer;

import de.fernunihagen.rtt.internal.util.ReversedList;

/**
 * Abstract test element container.
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class AbstractTestElementContainer extends AbstractTestElement
		implements ITestElementContainer {
	@Invar("null != $target")
	private final List<ITestElement> _elements;
	@Invar("null != $target")
	private Result _result;

	/**
	 * Creates a test container.
	 */
	public AbstractTestElementContainer() {
		this(Result.IGNORED, 0);
	}

	/**
	 * Creates a test container.
	 * 
	 * @param initSize
	 *            Initial size
	 */
	public AbstractTestElementContainer(final int initSize) {
		this(Result.IGNORED, initSize);
	}

	/**
	 * Creates a test container.
	 * 
	 * @param result
	 *            Initial result value
	 * @param initSize
	 *            Initial size
	 */
	@Pre("null != result")
	public AbstractTestElementContainer(final Result result, final int initSize) {
		_elements = new ArrayList<ITestElement>(initSize);
		_result = result;
	}

	@Override
	public void setTarget(Object target) {
		if (getParentContainer() == this) {
			return;
		}
		getParentContainer().setTarget(target);
	}

	@Override
	public Object getTarget() {
		if (getParentContainer() == this) {
			return null;
		}
		return super.getTarget();
	}

	@Override
	@Post
	public RunStage getStage() {
		if (getParentContainer() == this) {
			return RunStage.RUN;
		}
		return super.getStage();
	}

	@Pre
	@Post
	@Override
	public ITestElementContainer createChildContainer(final int initSize,
			final String label) {
		return new StructuralTestElementContainer(this, initSize, label);
	}

	@Pre
	@Post
	@Override
	public ITestElement createChild(final String label, final Result result) {
		return new StructuralTestElement(this, label, result);
	}

	@Override
	@Pre
	public void add(final ITestElement element) {
		_elements.add(element);
	}

	@Override
	@Post
	public ITestElement[] getChildren() {
		return _elements.toArray(new ITestElement[_elements.size()]);
	}

	/**
	 * Retrieves the count of children contained in container.
	 * 
	 * @return Count of children
	 */
	public int getChildrenCount() {
		return _elements.size();
	}

	@Override
	public boolean hasChildren() {
		return !_elements.isEmpty();
	}

	@Override
	@Post
	public long getElapsedTimeInTicks() {
		// FIXME: not efficient but necessary to ensure correct display of
		// time elapsed
		long current = super.getElapsedTimeInTicks();

		if (hasChildren()) {
			long elapsed = 0;

			for (ITestElement element : getChildren()) {
				elapsed += element.getElapsedTimeInTicks();
			}

			if (elapsed > current) {
				return elapsed;
			}
		}

		return current;
	}

	@Override
	@Post
	public Result getTestResult() {
		Result result = _result;

		for (ITestElement element : getChildren()) {
			result = result.combine(element.getTestResult());
		}

		return result;
	}

	@Override
	@Post
	public ProgressState getProgressState(final boolean includeChilds) {
		ProgressState state = super.getProgressState(false);

		if (includeChilds) {
			for (ITestElement element : getChildren()) {
				state = state.combine(element.getProgressState(includeChilds));
			}
		}

		return state;
	}

	@Override
	@Pre
	public void setProgressState(final ProgressState state) {
		super.setProgressState(state);
	}

	@Override
	protected boolean reset() {
		setTestResult(Result.IGNORED);

		final Iterator<ITestElement> reversed = ReversedList
				.reversed(_elements).iterator();

		while (reversed.hasNext()) {
			final ITestElement element = reversed.next();
			if (element instanceof AbstractTestElement && element != this) {
				((AbstractTestElement) element).reset();
			}
			reversed.remove();
		}

		return super.reset();
	}

	@Override
	@Pre
	public void setTestResult(final Result result) {
		_result = result;
	}

	@Override
	@Post("null != $return")
	public String toString() {
		if (getParentContainer() != this) {
			return super.toString();
		}

		return String.format("[/L %s]", getLabel());
	}
}
