package de.fernunihagen.rtt.framework.model;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;

import de.fernunihagen.rtt.framework.ITestActor;
import de.fernunihagen.rtt.framework.ITestElementContainer;

/**
 * Exception test element.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class ExceptionTestElement extends TestElement {
	@Invar("null != $target")
	private final Throwable _exception;
	@Invar("null != $target")
	private final ITestActor _originator;

	/**
	 * Creates a test exception element.
	 * 
	 * @param parent
	 *            Parent container
	 * @param originator
	 *            Originator
	 * @param exception
	 *            Exception
	 */
	@Pre("null != parent && null != originator && null != exception")
	public ExceptionTestElement(final ITestElementContainer parent,
			final ITestActor originator, final Throwable exception) {
		super(parent);
		_exception = exception;
		_originator = originator;

		parent.add(this);
		setProgressState(ProgressState.COMPLETED);
	}

	@Override
	public ITestActor getTarget() {
		return _originator;
	}

	@Override
	@Post("null != $return")
	public String toString() {
		final String message = String.format("[%s (/X %s) : %s]",
				getParentContainer(), _originator.getName(), _exception
						.getLocalizedMessage());
		return message;
	}

	@Override
	@Post
	public Result getTestResult() {
		return Result.ERROR;
	}

	private Throwable getRootException(final List<IStatus> statuses) {
		Throwable lowLevel = _exception;

		if (lowLevel instanceof CoreException
				&& null != ((CoreException) lowLevel).getStatus()) {
			IStatus root = ((CoreException) lowLevel).getStatus();
			lowLevel = root.getException();

			// check for nested exception
			if (null != lowLevel && lowLevel instanceof CoreException) {
				root = ((CoreException) lowLevel).getStatus();
			}

			lowLevel = root.getException();

			if (null != statuses) {
				statuses.add(root);

				final IStatus[] children = root.getChildren();
				for (IStatus status : children) {
					statuses.add(status);
				}
			}
		}

		if (lowLevel instanceof InvocationTargetException
				&& null != ((InvocationTargetException) lowLevel)
						.getTargetException()) {
			lowLevel = ((InvocationTargetException) lowLevel)
					.getTargetException();
		}

		return lowLevel;
	}

	/**
	 * Retrieves the stack frame pointing to exception reason.
	 * 
	 * @return Stack frame
	 */
	public String getStackFrame() {
		final StringBuffer buffer = new StringBuffer();
		final StackTraceElement[] elements;
		final Throwable lowLevel = getRootException(null);
		if (null != lowLevel) {
			elements = lowLevel.getStackTrace();
		} else {
			elements = new StackTraceElement[0];
		}

		for (int i = 0; i < elements.length; i++) {
			if (i != 0) {
				buffer.append("\n");
			}
			buffer.append(elements[i].toString());
		}

		return buffer.toString();
	}

	@Override
	@Post
	public String getLabel() {
		final StringBuffer buffer = new StringBuffer("Exception: ");
		final List<IStatus> statuses = new ArrayList<IStatus>();
		final Throwable lowLevel = getRootException(statuses);

		for (int i = 0; i < statuses.size(); i++) {
			if (i != 0) {
				buffer.append("\n");
			}
			buffer.append(statuses.get(i).getMessage());
		}

		if (null != lowLevel) {
			buffer.append("\n");
			buffer.append(lowLevel.getClass().getSimpleName());

			String message = lowLevel.getLocalizedMessage();

			if (null != message) {
				buffer.append(": ");
				buffer.append(message);
			}
		}

		return buffer.toString();
	}
}
