package de.fernunihagen.rtt.framework.model;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;

import de.fernunihagen.rtt.framework.ITestActorState;
import de.fernunihagen.rtt.framework.ITestElementContainer;

/**
 * Test run context.
 * 
 * @author Osama El Hosami
 */
@Contract
public class TestContext extends AbstractTestContext {

	@Invar("null != $target")
	private final ITestActorState _state;
	@Invar("null != $target")
	private final ITestElementContainer _parent;

	/**
	 * Creates a test run context.
	 * 
	 * @param state
	 *            Test actor state object
	 * @param parent
	 *            Parent test context
	 * @param result
	 *            Initial result value
	 */
	@Pre("null != state && null != parent && null != result")
	public TestContext(final ITestActorState state,
			final ITestElementContainer parent, final Result result) {
		super(result);
		_state = state;
		_parent = parent;

		parent.add(this);
	}

	/**
	 * Creates a test run context.
	 * 
	 * @param state
	 *            Test actor state object
	 * @param parent
	 *            Parent test context
	 */
	@Pre("null != state && null != parent")
	public TestContext(final ITestActorState state,
			final ITestElementContainer parent) {
		this(state, parent, Result.NONE);
	}

	@Override
	protected boolean reset() {
		getState().getOriginator().getActivator().stop(this);
		getState().reset();
		return super.reset();
	}

	@Override
	@Post
	public ITestElementContainer getParentContainer() {
		return _parent;
	}

	@Override
	@Post
	public ITestActorState getState() {
		return _state;
	}

	private void reportError(final String message, final Throwable exception)
			throws CoreException {
		getRTTCorePlugin().reportError(message, exception);
	}

	/**
	 * Activates context.
	 * 
	 * @param pm
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	public void activate(final IProgressMonitor pm) throws CoreException {
		try {
			final SubMonitor monitor = SubMonitor.convert(pm,
					"Activating test actor", 2);
			setProgressState(ProgressState.NOT_STARTED);
			if (getProgressState(true) != ProgressState.NOT_STARTED) {
				reportError(null, new IllegalStateException());
			}

			final ITestElementContainer init = createChildContainer(1,
					"Initialization");
			init.setProgressState(ProgressState.RUNNING);
			getState().initialize(init, monitor.newChild(1));
			init.setProgressState(ProgressState.COMPLETED);

			getState().getOriginator().getActivator().start(this,
					monitor.newChild(1));
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	/**
	 * Deactivates context.
	 */
	public void deactivate() {
		setProgressState(ProgressState.STOPPED);
		setProgressState(ProgressState.NOT_STARTED);
	}

	/**
	 * Runs test context.
	 * 
	 * @param pm
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @return true if execution is completed, otherwise context has pending
	 *         actions.
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	public boolean run(IProgressMonitor pm) throws CoreException {
		setProgressState(ProgressState.RUNNING);
		getState().getOriginator().apply(this, pm);
		setProgressState(ProgressState.COMPLETED);

		return !getState().isActive();
	}
}