package de.fernunihagen.rtt.framework.repository;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;

import org.eclipse.core.runtime.IPath;

/**
 * A diff describes differences between two or more model objects.
 * 
 * @author Osama El Hosami
 */
@Contract
public interface IDiff {
	/**
	 * Diff kind constant (bit mask) indicating that the resource has not been
	 * changed in any way.
	 */
	static final int NO_CHANGE = 0;

	/**
	 * Diff kind constant (bit mask) indicating that the resource has been added
	 * to its parent. That is, one that appears in the "after" state, not in the
	 * "before" one.
	 */
	static final int ADD = 0x1;

	/**
	 * Diff kind constant (bit mask) indicating that the resource has been
	 * removed from its parent. That is, one that appears in the "before" state,
	 * not in the "after" one.
	 */
	static final int REMOVE = 0x2;

	/**
	 * Diff kind constant (bit mask) indicating that the resource has been
	 * changed. That is, one that appears in both the "before" and "after"
	 * states.
	 */
	static final int CHANGE = 0x4;

	/**
	 * Change constant (bit mask) indicating that the content of the object has
	 * changed.
	 */
	static final int CONTENT = 0x100;

	/**
	 * Change constant (bit mask) indicating that the object was moved from
	 * another location. The location in the "before" state can be retrieved
	 * using <code>getFromPath()</code>.
	 */
	static final int MOVE_FROM = 0x200;

	/**
	 * Change constant (bit mask) indicating that the object was moved to
	 * another location. The location in the new state can be retrieved using
	 * <code>getToPath()</code>.
	 */
	static final int MOVE_TO = 0x400;

	/**
	 * Change constant (bit mask) indicating that the object was copied from
	 * another location. The location in the "before" state can be retrieved
	 * using <code>getFromPath()</code>.
	 */
	static final int COPY_FROM = 0x800;

	/**
	 * Change constant (bit mask) indicating that the object has been replaced
	 * by another at the same location (i.e., the object has been deleted and
	 * then added).
	 */
	static final int REPLACE = 0x1000;

	/**
	 * Returns the full, absolute path of the object to which the diff applies
	 * with respect to the model root.
	 * <p>
	 * Note: the returned path never has a trailing separator.
	 * </p>
	 * 
	 * @return the full, absolute path of this diff
	 */
	@Post("null != $return")
	IPath getPath();

	/**
	 * Returns the kind of this diff. Normally, one of <code>ADDED</code>,
	 * <code>REMOVED</code>, <code>CHANGED</code>.
	 * 
	 * @return the kind of this diff
	 */
	int getKind();

	/**
	 * Return a string that describes the difference represented by this node.
	 * 
	 * @return a string that describes the difference represented by this node
	 */
	@Post("null != $return")
	String toDiffString();

	/**
	 * Returns flags which describe in more detail how a object has been
	 * affected.
	 * <p>
	 * The following codes (bit masks) are used when kind is <code>CHANGE</code>
	 * , and also when the object is involved in a move:
	 * <ul>
	 * <li><code>CONTENT</code> - The bytes contained by the resource have been
	 * altered.</li>
	 * <li><code>REPLACE</code> - The object was deleted (either by a delete or
	 * move), and was subsequently re-created (either by a create, move, or
	 * copy).</li>
	 * </ul>
	 * The following code is only used if kind is <code>REMOVE</code> (or
	 * <code>CHANGE</code> in conjunction with <code>REPLACE</code>):
	 * <ul>
	 * <li><code>MOVE_TO</code> - The object has moved. <code>getToPath</code>
	 * will return the path of where it was moved to.</li>
	 * </ul>
	 * The following code is only used if kind is <code>ADD</code> (or
	 * <code>CHANGE</code> in conjunction with <code>REPLACE</code>):
	 * <ul>
	 * <li><code>MOVE_FROM</code> - The object has moved.
	 * <code>getFromPath</code> will return the path of where it was moved from.
	 * </li>
	 * <li><code>COPY_FROM</code> - The object has copied.
	 * <code>getFromPath</code> will return the path of where it was copied
	 * from.</li>
	 * </ul>
	 * A simple move operation would result in the following diff information.
	 * If a object is moved from A to B (with no other changes to A or B), then
	 * A will have kind <code>REMOVE</code>, with flag <code>MOVE_TO</code>, and
	 * <code>getToPath</code> on A will return the path for B. B will have kind
	 * <code>ADD</code>, with flag <code>MOVE_FROM</code>, and
	 * <code>getFromPath</code> on B will return the path for A. B's other flags
	 * will describe any other changes to the resource, as compared to its
	 * previous location at A.
	 * </p>
	 * <p>
	 * Note that the move flags only describe the changes to a single object;
	 * they don't necessarily imply anything about the parent or children of the
	 * object. If the children were moved as a consequence of a subtree move
	 * operation, they will have corresponding move flags as well.
	 * </p>
	 * 
	 * @return the flags
	 */
	int getFlags();

	/**
	 * Returns the full path (in the "before" state) from which this resource
	 * (in the "after" state) was moved. This value is only valid if the
	 * <code>MOVE_FROM</code> change flag is set; otherwise, <code>null</code>
	 * is returned.
	 * <p>
	 * Note: the returned path never has a trailing separator.
	 * 
	 * @return a path, or <code>null</code>
	 */
	IPath getFromPath();

	/**
	 * Returns the full path (in the "after" state) to which this resource (in
	 * the "before" state) was moved. This value is only valid if the
	 * <code>MOVE_TO</code> change flag is set; otherwise, <code>null</code> is
	 * returned.
	 * <p>
	 * Note: the returned path never has a trailing separator.
	 * 
	 * @return a path, or <code>null</code>
	 */
	IPath getToPath();
}
