package de.fernunihagen.rtt.framework.repository;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IResource;

import org.eclipse.core.runtime.CoreException;

/**
 * Repository provider, manages instance of repositories, which implements the
 * interface {@link IRepository}.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class RepositoryProvider extends AbstractRepositoryProvider {

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				RepositoryProvider.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	@Invar("null != $target")
	private final Map<String, IRepository> _map;

	/**
	 * Creates a query engine manager.
	 */
	public RepositoryProvider() {
		_map = Collections.synchronizedMap(new HashMap<String, IRepository>());
	}

	@Override
	@Pre
	public void addRepository(final IRepository repository)
			throws CoreException {
		final IRepository oldRepository = _map.put(repository.getID(),
				repository);
		if (null != oldRepository) {
			final String message = String.format(
					"Previously registered repository('%s') '%s' has been replaced"
							+ " by '%s'", repository.getID(), oldRepository
							.getName(), repository.getName());
			getLogger().debug(message);
		}
	}

	@Override
	@Pre
	@Post
	public IRepository getRepository(final IResource resource)
			throws CoreException {
		synchronized (_map) {
			for (Entry<String, IRepository> entry : _map.entrySet()) {
				final IRepository repository = entry.getValue();
				if (repository.getSubscriber().isSupervised(resource)) {
					return repository;
				}
			}
		}
		reportError("Repository of resource '"
				+ resource.getLocation().toPortableString()
				+ "' could not been found.");
		return null;
	}

	@Override
	@Pre
	@Post
	public IRepository getRepository(final String id) throws CoreException {
		final IRepository repository = _map.get(id);
		if (null == repository) {
			reportError("Repository with id '" + id + "' could not been found.");
		}
		return repository;
	}

	@Override
	public boolean isSupervised(final IResource resource) throws CoreException {
		synchronized (_map) {
			for (Entry<String, IRepository> entry : _map.entrySet()) {
				final IRepository repository = entry.getValue();
				if (repository.getSubscriber().isSupervised(resource)) {
					return true;
				}
			}
		}
		return false;
	}

	@Override
	public void removeRepository(final IRepository repository) {
		if (null != repository) {
			final IRepository removed = _map.remove(repository.getID());
			getLogger().assertLog(
					removed == repository,
					"Unrequested repository '" + repository.getID()
							+ "' has been removed");
		}
	}

	@Override
	public IRepository[] getRepositories() {
		return _map.values().toArray(new IRepository[_map.values().size()]);
	}
}
