package de.fernunihagen.rtt.framework.repository.team;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IProject;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;

import org.eclipse.team.core.ProjectSetCapability;
import org.eclipse.team.core.ProjectSetSerializationContext;
import org.eclipse.team.core.RepositoryProviderType;
import org.eclipse.team.core.TeamException;

import de.fernunihagen.rtt.framework.repository.IRepository;
import de.fernunihagen.rtt.framework.repository.ISubscriber;

import de.fernunihagen.rtt.framework.util.IOverwriteQuery;
import java.util.Arrays;

/**
 * Abstract team repository.
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class TeamRepository implements IRepository {

	@Invar("null != $target")
	private final RepositoryProviderType _providerType;

	/**
	 * Retrieves the repository provider type.
	 * 
	 * @return The repository provider type
	 */
	@Post("null != $return")
	protected final RepositoryProviderType getProviderType() {
		return _providerType;
	}

	/**
	 * Retrieves the object instance used for project serialization and
	 * deserialization.
	 * 
	 * @return The project set capability instance
	 */
	@Post("null != $return")
	protected final ProjectSetCapability getProjectSetCapability() {
		return getProviderType().getProjectSetCapability();
	}

	/**
	 * Creates a team repository for given provider type
	 * 
	 * @param providerType
	 *            Repository provider type
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             <ul>
	 */
	private TeamRepository(final RepositoryProviderType providerType)
			throws CoreException {

		if (null == providerType
				|| null == providerType.getProjectSetCapability()) {
			throw new TeamException("Repository type could not be found.");
		}

		_providerType = providerType;
	}

	/**
	 * Creates a team repository for given repository provider type identifier
	 * 
	 * @param id
	 *            Repository provider type id
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             <ul>
	 */
	@Pre("null != id")
	protected TeamRepository(final String id) throws CoreException {
		this(RepositoryProviderType.getProviderType(id));
	}

	@Override
	@Pre
	@Post
	public String[] getReferences(final IProject[] projects,
			final IProgressMonitor pm) throws CoreException {

		try {
			final SubMonitor monitor = SubMonitor.convert(pm);

			final ProjectSetCapability serializer = getProjectSetCapability();

			ProjectSetCapability.ensureBackwardsCompatible(_providerType,
					serializer);

			return serializer.asReference(projects,
					new ProjectSetSerializationContext(), monitor);
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Override
	@Pre
	@Post
	public IProject[] checkout(final String[] references,
			final IOverwriteQuery query, final IProgressMonitor pm)
			throws CoreException {

		try {
			final SubMonitor monitor = SubMonitor.convert(pm);

			final ProjectSetCapability serializer = getProjectSetCapability();

			ProjectSetCapability.ensureBackwardsCompatible(_providerType,
					serializer);

			return serializer.addToWorkspace(references,
					new ProjectSetSerializationContext() {
						@Override
						public IProject[] confirmOverwrite(
								final IProject[] projects) throws TeamException {
							final List<IProject> toOverwrite = new ArrayList<IProject>(
									projects.length);
							query: for (IProject project : projects) {
								if (!project.exists()) {
									toOverwrite.add(project);
									continue;
								}

								switch (query.queryOverwrite(project
										.getLocation().toPortableString())) {
								case CANCEL:
									throw TeamException
											.asTeamException(new InvocationTargetException(
													new OperationCanceledException()));
								case YES:
									toOverwrite.add(project);
									break;
								case ALL:
									toOverwrite.addAll(Arrays.asList(projects));
									break query;
								case NO_ALL:
									toOverwrite.clear();
									break query;
								case NO:
								default:
									break;
								}
							}
							return toOverwrite.toArray(new IProject[toOverwrite
									.size()]);
						}
					}, monitor);
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Override
	@Post
	public ISubscriber getSubscriber() throws CoreException {
		return new TeamSubscriber(getProviderType().getSubscriber());
	}

	@Override
	public String getID() {
		return getProviderType().getID();
	}

	@Override
	public String getName() {
		return getProviderType().getSubscriber().getName();
	}

	@Override
	public int hashCode() {
		return getProviderType().hashCode();
	}

	@Override
	public String toString() {
		return getProviderType().toString();
	}

	@Override
	public boolean equals(final Object other) {
		if (other == this) {
			return true;
		}
		if (null == other) {
			return false;
		}
		return other.equals(getProviderType());
	}
}
