package de.fernunihagen.rtt.framework.testadapter;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import java.lang.reflect.Method;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;

import de.fernunihagen.rtt.framework.ITestContext;
import de.fernunihagen.rtt.framework.ITestElement;
import de.fernunihagen.rtt.framework.ITestElementContainer;
import de.fernunihagen.rtt.framework.ITestRunSession;

import de.fernunihagen.rtt.framework.ITestElement.ProgressState;
import de.fernunihagen.rtt.framework.ITestElement.Result;

import de.fernunihagen.rtt.framework.model.ExceptionTestElement;

import de.fernunihagen.rtt.framework.testadapter.RefactoringTestRunState.TestAdapter;

/**
 * Reflective refactoring test adapter.
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class ReflectiveRefactoringTestAdapter extends
		RefactoringTestAdapter {

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				ReflectiveRefactoringTestAdapter.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Create a reflective refactoring test adapter.
	 */
	public ReflectiveRefactoringTestAdapter() {
		super();
	}

	private void reportError(final CoreException[] exceptions,
			final String message) throws CoreException {
		getRTTCorePlugin().reportError(exceptions, message);
	}

	/**
	 * Tests if test of test case is accepted.
	 * 
	 * @param test
	 *            Test
	 * @param testcase
	 *            Test case
	 * 
	 * @return true if test is accepted
	 */
	@Pre("null != test && null != testcase")
	protected boolean accept(final Method test, final Class<?> testcase) {
		return true;
	}

	@Override
	@Pre
	public void start(final ITestContext context, final IProgressMonitor pm)
			throws CoreException {

		try {
			// FIXME: make configurable
			final boolean fShouldContinueOnError = false;
			final RefactoringTestRunState state = getState(context);
			final List<CoreException> errors = new ArrayList<CoreException>();
			final RefactoringTestDescription description = state
					.getDescription();
			final String[] testCases = description.getTestCases();
			final ITestElementContainer container = context
					.createChildContainer(testCases.length, "Test cases");

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Loading test case(s)", testCases.length);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format(
						"Loading test case(s) (%s)", testCases.length);
				getLogger().debug(message);
			}

			container.setProgressState(ProgressState.RUNNING);

			for (String name : testCases) {
				if (monitor.isCanceled()) {
					throw new OperationCanceledException();
				}

				monitor.subTask(name);

				final ITestElementContainer element = container
						.createChildContainer(0, name);

				if (getLogger().isDebugEnabled()) {
					final String message = String.format(
							"Analyzing test case(s) (%s)", name);
					getLogger().debug(message);
				}

				element.setProgressState(ProgressState.RUNNING);

				if (errors.isEmpty() || fShouldContinueOnError) {
					element.setTestResult(Result.OK);

					try {
						final Class<?> testcase = description.loadClass(name);
						final Method[] candidates = RefactoringTestDescription
								.getTests(new Class<?>[] { testcase });

						for (Method test : candidates) {
							final ITestElement testElement;
							final boolean fIsAccept = accept(test, testcase);
							final String testName = String.format("%s (%08X)",
									test.getName(), test.hashCode());

							if (fIsAccept) {
								state.addTest(test, testcase);
								testElement = element.createChild(testName,
										Result.NONE);
							} else {
								testElement = element.createChild(testName,
										Result.IGNORED);
							}

							if (getLogger().isDebugEnabled()) {
								final String message = String.format(
										"%s (%s;%s;%.3fs)", testElement,
										testElement.getTestResult(),
										testElement.getProgressState(true),
										element.getElapsedTimeInSeconds());
								getLogger().debug(message);
							}
						}

					} catch (OperationCanceledException e) {
						element.setProgressState(ProgressState.STOPPED);
						throw e;
					} catch (CoreException e) {
						getLogger().trace("Unexpected", e);
						element.setTestResult(Result.ERROR);
						errors.add(e);
					}
				}
				// else ignored notifications

				element.setProgressState(ProgressState.COMPLETED);

				if (getLogger().isDebugEnabled()) {
					final String message = String.format("%s (%s;%s;%.3fs)",
							element, element.getTestResult(), element
									.getProgressState(true), element
									.getElapsedTimeInSeconds());
					getLogger().debug(message);
				}

				monitor.worked(1);
			}

			container.setProgressState(ProgressState.COMPLETED);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format("Load done (%s;%s;%.3fs)",
						container.getTestResult(), container
								.getProgressState(true), container
								.getElapsedTimeInSeconds());
				getLogger().debug(message);
			}

			if (!errors.isEmpty()) {
				reportError(errors.toArray(new CoreException[errors.size()]),
						"Error ocurred while loading refactoring test case(s)");
			}

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	/**
	 * Runs a test adapter.
	 * 
	 * @param testadapter
	 *            Test adapter
	 * @param element
	 *            Element container
	 * @param monitor
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @return true if execution is completed, otherwise test adapter has to be
	 *         executed again.
	 * 
	 * @throws CoreException
	 */
	@Pre("null != element && null != testadapter")
	protected abstract boolean run(final TestAdapter testadapter,
			final ITestElementContainer element, IProgressMonitor monitor)
			throws CoreException;

	@Override
	@Pre
	@Post
	public abstract RefactoringTestRunState create(ITestRunSession session,
			IProgressMonitor monitor) throws CoreException;

	@Override
	@Pre
	public void apply(final ITestContext context, final IProgressMonitor pm)
			throws CoreException {
		try {
			final RefactoringTestRunState state = getState(context);

			if (state.hasNext()) {
				// FIXME make configurable
				final boolean fShouldContinueOnError = false;
				boolean fSatisfied = context.getTestResult().isSuccess();
				final TestAdapter testadapter = state.nextTestAdapter(context
						.getSession());
				final ITestElementContainer container = context;

				final SubMonitor monitor = SubMonitor.convert(pm,
						"Running test", 1);

				if (getLogger().isDebugEnabled()) {
					final String message = String.format("Running test %s",
							testadapter.toString());
					getLogger().debug(message);
				}

				final ITestElementContainer element = container
						.createChildContainer(0, testadapter.toString());

				element.setTarget(testadapter.getDescription());
				monitor.subTask(testadapter.toString());

				element.setProgressState(ProgressState.RUNNING);

				if (fSatisfied || state.isStalled() || fShouldContinueOnError) {
					try {
						state.setStalled(!run(testadapter, element, monitor
								.newChild(1)));
					} catch (OperationCanceledException e) {
						element.setProgressState(ProgressState.STOPPED);
						throw e;
					} catch (CoreException e) {
						new ExceptionTestElement(element, this, e);
						getLogger().trace("Unexpected", e);
						state.setStalled(false);
						state.addError(e);
					}

				}
				// else ignored notifications

				element.setProgressState(ProgressState.COMPLETED);

				if (getLogger().isDebugEnabled()) {
					final String message = String.format("%s (%s;%s;%.3fs)",
							element, element.getTestResult(), element
									.getProgressState(true), element
									.getElapsedTimeInSeconds());
					getLogger().debug(message);
				}

				getLogger().debug("Run done");

				state.seekNext();

			} else if (state.hasError()) {
				final String message = "Failed to run test(s), least one test failed to run";
				state.reportError(message);
			}

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}
}
