package de.fernunihagen.rtt.internal;

import java.io.Closeable;

import java.util.Enumeration;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.spi.LoggerFactory;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.Plugin;

import com.tools.logging.PluginLogManager;

/**
 * An abstract log manager.
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class AbstractLogManager implements Closeable {
	@Invar("null != $target")
	private final Plugin _plugin;

	/**
	 * Creates a log manager.
	 * 
	 * @param plugin
	 *            Plug-in
	 */
	@Pre("null != plugin")
	protected AbstractLogManager(final Plugin plugin) {
		_plugin = plugin;
	}

	/**
	 * Retrieves the log manager owner.
	 * 
	 * @return Log manager owner
	 */
	@Post("null != $return")
	protected final Plugin getOwner() {
		return _plugin;
	}

	/**
	 * Retrieves the log manager owner identifier.
	 * 
	 * @return Log manager owner identifier
	 */
	@Post("null != $return")
	protected final String getOwnerId() {
		return getOwner().getBundle().getSymbolicName();
	}

	/**
	 * Retrieves the adaptee instance.
	 * 
	 * @return The adaptee instance
	 */
	@Post("null != $return")
	protected abstract PluginLogManager getAdaptee();

	/**
	 * Initializes the log manager.
	 * 
	 * @return True if successful otherwise fail
	 */
	public abstract boolean init();

	/**
	 * Tests if log manager has been closed
	 * 
	 * @return True if satisfied otherwise false
	 */
	public abstract boolean isClosed();

	/**
	 * Hooks into the log of a plug-in
	 * 
	 * @param plugin
	 *            The plug-in to hook for logging
	 * 
	 * @return true if succeeded otherwise false
	 */
	@Pre("null != plugin")
	protected final boolean hookPlugin(final Plugin plugin) {
		return getAdaptee().hookPlugin(plugin.getBundle().getSymbolicName(),
				plugin.getLog());
	}

	/**
	 * Unhooks a plug-in from this log manager. The Eclipse log API won't be
	 * channeled to this logging framework anymore.
	 * 
	 * @param plugin
	 *            The plug-in to unhook for logging
	 * 
	 * @return true if succeeded otherwise false
	 */
	@Pre("null != plugin")
	protected final boolean unhookPlugin(final Plugin plugin) {
		return getAdaptee().unHookPlugin(plugin.getBundle().getSymbolicName());
	}

	/**
	 * Checks if this log manager is disabled for this level.
	 * 
	 * @param level
	 *            Level value
	 * 
	 * @return boolean true if it is disabled
	 */
	public boolean isDisabled(final int level) {
		return getAdaptee().isDisabled(level);
	}

	/**
	 * Enable logging for logging requests with level l or higher. By default
	 * all levels are enabled.
	 * 
	 * @param level
	 *            Level object
	 */
	@Pre("null != level")
	public void setThreshold(final Level level) {
		getAdaptee().setThreshold(level);
	}

	/**
	 * The string version of setThreshold(Level level)
	 * 
	 * @param level
	 *            level string
	 */
	@Pre("null != level")
	public void setThreshold(final String level) {
		getAdaptee().setThreshold(level);
	}

	/**
	 * Get the repository-wide threshold.
	 * 
	 * @return Level
	 */
	@Post("null != $return")
	public Level getThreshold() {
		return getAdaptee().getThreshold();
	}

	/**
	 * Returns a new logger instance named as the first parameter using the
	 * default factory. If a logger of that name already exists, then it will be
	 * returned. Otherwise, a new logger will be instantiated and then linked
	 * with its existing ancestors as well as children.
	 * 
	 * @param name
	 *            Logger name
	 * 
	 * @return Logger
	 */
	@Pre("null != name")
	@Post("null != $return")
	public Logger getLogger(final String name) {
		return getAdaptee().getLogger(name);
	}

	/**
	 * The same as getLogger(String name) but using a factory instance instead
	 * of a default factory.
	 * 
	 * @param name
	 *            logger name
	 * @param factory
	 *            factory instance
	 * @return Logger
	 */
	@Pre("null != name && null != factory")
	@Post("null != $return")
	public Logger getLogger(final String name, final LoggerFactory factory) {
		return getAdaptee().getLogger(name, factory);
	}

	/**
	 * Returns the root of this hierarchy.
	 * 
	 * @return Logger
	 */
	@Post("null != $return")
	public Logger getRootLogger() {
		return getAdaptee().getRootLogger();
	}

	/**
	 * Checks if this logger exists.
	 * 
	 * @return Logger
	 */
	@Pre("null != name")
	@Post("null != $return")
	public Logger exists(final String name) {
		return getAdaptee().exists(name);
	}

	/**
	 * Returns all the loggers in this manager.
	 * 
	 * @return Enumeration logger enumeration
	 */
	@SuppressWarnings("unchecked")
	@Post("null != $return")
	public Enumeration<Logger> getCurrentLoggers() {
		return getAdaptee().getCurrentLoggers();
	}

	/**
	 * Resets configuration values to its defaults.
	 */
	public void resetConfiguration() {
		getAdaptee().resetConfiguration();
	}
}
