package de.fernunihagen.rtt.internal;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Plugin;
import org.eclipse.core.runtime.Status;

import org.osgi.framework.BundleContext;
import org.osgi.framework.Filter;
import org.osgi.framework.FrameworkUtil;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;

import org.osgi.util.tracker.ServiceTracker;

import de.fernunihagen.rtt.framework.ITestRunListenerRegistry;

import de.fernunihagen.rtt.framework.oracle.IOracleProvider;
import de.fernunihagen.rtt.framework.oracle.IOracleRunner;

import de.fernunihagen.rtt.framework.repository.IRepositoryProvider;

import de.fernunihagen.rtt.framework.testadapter.IRefactoringTestAdapterProvider;
import de.fernunihagen.rtt.framework.testadapter.IRefactoringTestRunner;

/**
 * The activator class controls the plug-in life cycle
 */
@Contract
public final class RTTCorePlugin extends Plugin {

	// The plug-in ID
	public static final String PLUGIN_ID = "de.fernunihagen.rtt";

	// The shared instance
	private static RTTCorePlugin plugin;

	private static final String _LOG_PROPERTIES_FILE = "logger.properties";
	private static final long SERVICE_TIMEOUT = 3000;

	private PluginAdapter<RTTCorePlugin> _adapter;
	@Invar("null != $target")
	private Filter _activeTestRunListenerRegistry;
	@Invar("null != $target")
	private Filter _activeRepositoryProvider;
	@Invar("null != $target")
	private Filter _activeOracleProvider;
	@Invar("null != $target")
	private Filter _activeRefactoringAdapterProvider;
	@Invar("null != $target")
	private Filter _activeOracleRunner;
	@Invar("null != $target")
	private Filter _activeRefactoringRunner;

	private ServiceTracker _testRunListenerRegistry;
	private ServiceTracker _oracleRunner;
	private ServiceTracker _refactoringRunner;
	private ServiceTracker _repositoryProvider;
	private ServiceTracker _oracleProvider;
	private ServiceTracker _refactoringAdapterProvider;
	private volatile boolean _fInit;

	/**
	 * The constructor.
	 */
	public RTTCorePlugin() {
		_fInit = false;
		_adapter = null;

		setActiveTestRunListenerRegistry("de.fernunihagen.rtt.runner.testrunlistener.registry");
		setActiveRepositoryProvider("de.fernunihagen.rtt.repository.provider");
		setActiveOracleProvider("de.fernunihagen.rtt.oracle.provider");
		setActiveRefactoringTestAdapterProvider("de.fernunihagen.rtt.testadapter.provider");
		setActiveOracleRunner("de.fernunihagen.rtt.oracle.runner");
		setActiveRefactoringTestRunner("de.fernunihagen.rtt.testadapter.runner");
	}

	/**
	 * Sets the active test run listener registry.
	 * 
	 * @param name
	 *            Test run listener registry name
	 */
	@Pre("null != name")
	public synchronized void setActiveTestRunListenerRegistry(final String name) {
		_activeTestRunListenerRegistry = filterByComponentName(name);
	}

	@Post("null != $return")
	private synchronized Filter getActiveTestRunListenerRegistry() {
		return _activeTestRunListenerRegistry;
	}

	/**
	 * Sets the active repository provider.
	 * 
	 * @param name
	 *            Repository provider name
	 */
	@Pre("null != name")
	public synchronized void setActiveRepositoryProvider(final String name) {
		_activeRepositoryProvider = filterByComponentName(name);
	}

	@Post("null != $return")
	private synchronized Filter getActiveRepositoryProvider() {
		return _activeRepositoryProvider;
	}

	/**
	 * Sets the active oracle provider.
	 * 
	 * @param name
	 *            Oracle provider name
	 */
	@Pre("null != name")
	public synchronized void setActiveOracleProvider(final String name) {
		_activeOracleProvider = filterByComponentName(name);
	}

	@Post("null != $return")
	private synchronized Filter getActiveOracleProvider() {
		return _activeOracleProvider;
	}

	/**
	 * Sets the active refactoring test adapter provider.
	 * 
	 * @param name
	 *            Refactoring test adapter provider name
	 */
	@Pre("null != name")
	public synchronized void setActiveRefactoringTestAdapterProvider(
			final String name) {
		_activeRefactoringAdapterProvider = filterByComponentName(name);
	}

	@Post("null != $return")
	private synchronized Filter getActiveRefactoringTestAdapterProvider() {
		return _activeRefactoringAdapterProvider;
	}

	/**
	 * Sets the active oracle runner.
	 * 
	 * @param name
	 *            Oracle runner name
	 */
	@Pre("null != name")
	public synchronized void setActiveOracleRunner(final String name) {
		_activeOracleRunner = filterByComponentName(name);
	}

	@Post("null != $return")
	private synchronized Filter getActiveOracleRunner() {
		return _activeOracleRunner;
	}

	/**
	 * Sets the active refactoring test runner.
	 * 
	 * @param name
	 *            Refactoring test runner name
	 */
	@Pre("null != name")
	public synchronized void setActiveRefactoringTestRunner(final String name) {
		_activeRefactoringRunner = filterByComponentName(name);
	}

	@Post("null != $return")
	private synchronized Filter getActiveRefactoringTestRunner() {
		return _activeRefactoringRunner;
	}

	@Pre("null != name")
	@Post("null != $return")
	private Filter filterByComponentName(final String name) {
		try {
			return FrameworkUtil.createFilter(String.format("(%s=%s)",
					"component.name", name));
		} catch (InvalidSyntaxException e) {
			// unexpected
		}
		return null;
	}

	@Pre("null != tracker && null != filter")
	private ServiceReference findServiceReferenceByFilter(
			final ServiceTracker tracker, final Filter filter) {
		final ServiceReference[] references = tracker.getServiceReferences();
		for (ServiceReference reference : references) {
			if (filter.match(reference)) {
				return reference;
			}
		}

		getAdapter().logWarning(
				"No matching service reference, using first registered.");

		return tracker.getServiceReference();
	}

	@Pre("null != tracker && null != filter && null != clazz")
	private <T> T findServiceByFilter(final ServiceTracker tracker,
			final Filter filter, final Class<T> clazz) {
		initialize();
		try {
			tracker.waitForService(SERVICE_TIMEOUT);
		} catch (InterruptedException e) {
		}
		return clazz.cast(tracker.getService(findServiceReferenceByFilter(
				tracker, filter)));
	}

	@Override
	@Pre("$this.PLUGIN_ID == context.getBundle().getSymbolicName()")
	public void start(final BundleContext context) throws Exception {
		try {
			super.start(context);
			plugin = this;
			_adapter = new PluginAdapter<RTTCorePlugin>(this,
					_LOG_PROPERTIES_FILE);
			_testRunListenerRegistry = new ServiceTracker(context,
					ITestRunListenerRegistry.class.getName(), null);
			_repositoryProvider = new ServiceTracker(context,
					IRepositoryProvider.class.getName(), null);
			_oracleProvider = new ServiceTracker(context, IOracleProvider.class
					.getName(), null);
			_refactoringAdapterProvider = new ServiceTracker(context,
					IRefactoringTestAdapterProvider.class.getName(), null);
			_oracleRunner = new ServiceTracker(context, IOracleRunner.class
					.getName(), null);
			_refactoringRunner = new ServiceTracker(context,
					IRefactoringTestRunner.class.getName(), null);

			getAdapter().logInfo(PLUGIN_ID + " started");
		} catch (Exception e) {
			getLog().log(new Status(IStatus.ERROR, PLUGIN_ID, null, e));
		}
	}

	/**
	 * Initializes service trackers.
	 */
	public void initialize() {
		if (!_fInit) {
			synchronized (this) {
				if (!_fInit) {
					_testRunListenerRegistry.open();
					_repositoryProvider.open();
					_oracleProvider.open();
					_refactoringAdapterProvider.open();
					_oracleRunner.open();
					_refactoringRunner.open();
					_fInit = true;
				}
			}
		}
	}

	@Override
	public void stop(final BundleContext context) throws Exception {
		_adapter.logInfo(PLUGIN_ID + " stopped");
		_adapter.close();
		_adapter = null;

		synchronized (this) {
			if (_fInit) {
				_refactoringRunner.close();
				_oracleRunner.close();
				_refactoringAdapterProvider.close();
				_oracleProvider.close();
				_repositoryProvider.close();
				_testRunListenerRegistry.close();
			}
		}

		plugin = null;
		super.stop(context);
	}

	/**
	 * Returns the shared instance
	 * 
	 * @return the shared instance
	 */
	public static RTTCorePlugin getDefault() {
		return plugin;
	}

	/**
	 * Returns the adapted shared instance
	 * 
	 * @return the adapted shared instance
	 */
	public static PluginAdapter<? extends Plugin> getRTTCorePlugin() {
		return getDefault().getAdapter();
	}

	/**
	 * Retrieves the repository provider.
	 * 
	 * @return Repository provider
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If service unavailable</li>
	 *             <ul>
	 */
	@Post("null != $return")
	public IRepositoryProvider getRepositoryProvider() throws CoreException {
		IRepositoryProvider provider = null;
		if (null != _repositoryProvider) {
			provider = findServiceByFilter(_repositoryProvider,
					getActiveRepositoryProvider(), IRepositoryProvider.class);
		}
		if (null == provider) {
			getAdapter().reportError("Repository provider service unavailable");
		}
		return provider;
	}

	/**
	 * Retrieves the oracle provider.
	 * 
	 * @return Oracle provider
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If service unavailable</li>
	 *             <ul>
	 */
	@Post("null != $return")
	public IOracleProvider getOracleProvider() throws CoreException {
		IOracleProvider provider = null;
		if (null != _oracleProvider) {
			provider = findServiceByFilter(_oracleProvider,
					getActiveOracleProvider(), IOracleProvider.class);
		}
		if (null == provider) {
			getAdapter().reportError("Oracle provider service unavailable");
		}
		return provider;
	}

	/**
	 * Retrieves the refactoring test adapter provider.
	 * 
	 * @return Refactoring test adapter
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If service unavailable</li>
	 *             <ul>
	 */
	@Post("null != $return")
	public IRefactoringTestAdapterProvider getRefactoringTestAdapterProvider()
			throws CoreException {
		IRefactoringTestAdapterProvider provider = null;
		if (null != _refactoringAdapterProvider) {
			provider = findServiceByFilter(_refactoringAdapterProvider,
					getActiveRefactoringTestAdapterProvider(),
					IRefactoringTestAdapterProvider.class);
		}
		if (null == provider) {
			getAdapter().reportError(
					"Refactoring test adapter provider service unavailable");
		}
		return provider;
	}

	/**
	 * Retrieves the refactoring test runner.
	 * 
	 * @return Refactoring test runner
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If service unavailable</li>
	 *             <ul>
	 */
	@Post("null != $return")
	public IRefactoringTestRunner getRefactoringTestRunner()
			throws CoreException {
		IRefactoringTestRunner runner = null;
		if (null != _refactoringRunner) {
			runner = findServiceByFilter(_refactoringRunner,
					getActiveRefactoringTestRunner(),
					IRefactoringTestRunner.class);
		}
		if (null == runner) {
			getAdapter().reportError(
					"Refactoring test runner service unavailable");
		}
		return runner;
	}

	/**
	 * Retrieves the oracle runner.
	 * 
	 * @return Oracle runner
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If service unavailable</li>
	 *             <ul>
	 */
	@Post("null != $return")
	public IOracleRunner getOracleRunner() throws CoreException {
		IOracleRunner runner = null;
		if (null != _oracleRunner) {
			runner = findServiceByFilter(_oracleRunner,
					getActiveOracleRunner(), IOracleRunner.class);
		}
		if (null == runner) {
			getAdapter().reportError("Oracle runner service unavailable");
		}
		return runner;
	}

	/**
	 * Retrieves the test run listeners registry.
	 * 
	 * @return Test run listeners registry
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If service unavailable</li>
	 *             <ul>
	 */
	@Post("null != $return")
	public ITestRunListenerRegistry getTestRunListenerRegistry()
			throws CoreException {
		ITestRunListenerRegistry registry = null;
		if (null != _testRunListenerRegistry) {
			registry = findServiceByFilter(_testRunListenerRegistry,
					getActiveTestRunListenerRegistry(),
					ITestRunListenerRegistry.class);
		}
		if (null == registry) {
			getAdapter().reportError(
					"Test run listener registry service unavailable");
		}
		return registry;
	}

	private PluginAdapter<? extends Plugin> getAdapter() {
		if (null == _adapter) {
			return null;
		}

		_adapter.init();
		return _adapter;
	}
}
