package de.fernunihagen.rtt.internal.jdt;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.log4j.Logger;
import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;
import org.eclipse.core.resources.IProject;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;

import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaCore;

import org.eclipse.jdt.junit.JUnitCore;

import de.fernunihagen.rtt.framework.IElementCollector;
import de.fernunihagen.rtt.framework.IElementFilter;

/**
 * JUnit test case collector.
 * 
 * @author Osama El Hosami
 */
@Contract
public class JUnitTestCaseCollector implements IElementCollector<IType> {
	@Invar("null != $target")
	private final List<IType> _list;
	@Invar("null != $target")
	private final IElementFilter<IAdaptable> _filter;
	@Invar("null != $target")
	private final IProject[] _scope;

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				JUnitTestCaseCollector.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Create a JUnit test case collector, used to fill given list with matching
	 * elements. List has to be synchronized in case collector is used
	 * concurrently.
	 * 
	 * @param list
	 *            List
	 * @param filter
	 *            Filter
	 * @param scope
	 *            Projects in search scope
	 */
	@Pre("null != list && null != filter && null != scope")
	public JUnitTestCaseCollector(final List<IType> list,
			final IElementFilter<IAdaptable> filter, final IProject[] scope) {
		_list = list;
		_filter = filter;
		_scope = scope;
	}

	/**
	 * Creates a JUnit test case collector, used to collect matching elements.
	 * 
	 * @param filter
	 *            Filter
	 * @param scope
	 *            Projects in search scope
	 */
	@Pre("null != filter && null != scope")
	public JUnitTestCaseCollector(final IElementFilter<IAdaptable> filter,
			final IProject[] scope) {
		this(new ArrayList<IType>(), filter, scope);
	}

	/**
	 * Creates a JUnit test case collector.
	 * 
	 * @param scope
	 *            Projects in search scope
	 */
	@Pre("null != scope")
	public JUnitTestCaseCollector(final IProject[] scope) {
		this(IElementFilter.ALL, scope);
	}

	@Pre("null != types && null != monitor")
	private void collect(final IType[] types, final SubMonitor monitor)
			throws CoreException {

		monitor.beginTask("Accepting JUnit test case(s)", types.length);
		getLogger().debug("Accepting JUnit test case(s)");

		for (IType type : types) {
			final boolean fAccept = _filter.accept(type);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format("[T/ %s (%s)]", type
						.getElementName(), fAccept);
				getLogger().debug(message);
			}

			if (fAccept) {
				monitor.subTask(type.getElementName());
				_list.add(type);
			}

			monitor.worked(1);
		}

		getLogger().debug("Accept done");
	}

	@Override
	public void collect(final IProgressMonitor pm) throws CoreException {
		try {
			clear();

			final IProject[] projects = getScope();

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Collecting JUnit test case(s)", 3 * projects.length);
			getLogger().debug("Collecting JUnit test case(s)");

			for (IProject project : projects) {
				project.open(monitor.newChild(1));

				final boolean fIsJavaProject = project
						.hasNature(JavaCore.NATURE_ID);

				if (getLogger().isDebugEnabled()) {
					final String message = String.format("[P/ %s (%s)]",
							project.getName(), fIsJavaProject);
					getLogger().debug(message);
				}

				if (fIsJavaProject) {
					final IJavaProject javaProject = JavaCore.create(project);
					final IType[] types = JUnitCore.findTestTypes(javaProject,
							monitor.newChild(1));
					collect(types, monitor.newChild(1));
				} else {
					monitor.worked(2);
				}
			}

			if (getLogger().isDebugEnabled()) {
				final String message = String.format("Collection done (%d)",
						_list.size());
				getLogger().debug(message);
			}

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Override
	public List<? extends IType> getMatches() {
		return Collections.unmodifiableList(_list);
	}

	@Override
	public IProject[] getScope() {
		return _scope;
	}

	@Override
	public void clear() {
		_list.clear();
	}

}
