package de.fernunihagen.rtt.internal.jdt;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;

import org.eclipse.jdt.core.IJavaElement;

import org.eclipse.jdt.core.search.SearchEngine;
import org.eclipse.jdt.core.search.SearchMatch;
import org.eclipse.jdt.core.search.SearchParticipant;
import org.eclipse.jdt.core.search.SearchRequestor;

import de.fernunihagen.rtt.framework.IElementCollector;
import de.fernunihagen.rtt.framework.IElementFilter;

/**
 * Java element collector.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class JavaElementCollector extends SearchRequestor implements
		IElementCollector<IJavaElement> {

	@Invar("null != $target")
	private final List<IJavaElement> _list;
	@Invar("null != $target")
	private final IElementFilter<IAdaptable> _filter;
	@Invar("null != $target")
	private final JavaSearchPatternFactory _factory;

	/**
	 * Create a java element collector, used to fill given list with matching
	 * elements. List has to be synchronized in case collector is used
	 * concurrently.
	 * 
	 * @param list
	 *            List
	 * @param filter
	 *            Filter
	 * @param factory
	 *            Search pattern factory
	 */
	@Pre("null != list && null != filter && null != factory")
	public JavaElementCollector(final List<IJavaElement> list,
			final IElementFilter<IAdaptable> filter,
			final JavaSearchPatternFactory factory) {
		_list = list;
		_filter = filter;
		_factory = factory;
	}

	/**
	 * Creates a java element collector, used to collect matching elements.
	 * 
	 * @param filter
	 *            Filter
	 * @param factory
	 *            Search pattern factory
	 */
	@Pre("null != filter && null != factory")
	public JavaElementCollector(final IElementFilter<IAdaptable> filter,
			final JavaSearchPatternFactory factory) {
		this(new ArrayList<IJavaElement>(), filter, factory);
	}

	/**
	 * Creates a java element collector.
	 * 
	 * @param factory
	 *            Search pattern factory
	 */
	@Pre("null != factory")
	public JavaElementCollector(final JavaSearchPatternFactory factory) {
		this(IElementFilter.ALL, factory);
	}

	@Post("null != $return")
	@Override
	public List<? extends IJavaElement> getMatches() {
		return Collections.unmodifiableList(_list);
	}

	@Override
	@Pre("null != match")
	public void acceptSearchMatch(final SearchMatch match) throws CoreException {
		if (match.getAccuracy() == SearchMatch.A_ACCURATE
				&& !match.isInsideDocComment()) {
			final Object element = match.getElement();

			if (element instanceof IJavaElement
					&& _filter.accept((IJavaElement) element)) {
				_list.add((IJavaElement) element);
			}
		}
	}

	@Override
	public void collect(final IProgressMonitor monitor) throws CoreException {
		clear();

		SearchParticipant[] participants = new SearchParticipant[] { SearchEngine
				.getDefaultSearchParticipant() };

		new SearchEngine().search(_factory.create(), participants, _factory
				.getSearchScope(), this, monitor);
	}

	@Override
	@Post("null != $return")
	public IProject[] getScope() {
		final IJavaElement[] elements = _factory.getRegion();
		final Set<IProject> projects = new HashSet<IProject>();
		for (IJavaElement element : elements) {
			projects.add(element.getJavaProject().getProject());
		}
		return projects.toArray(new IProject[projects.size()]);
	}

	@Override
	public void clear() {
		_list.clear();
	}
}
