package de.fernunihagen.rtt.internal.jdt;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IProject;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.SubMonitor;

import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;

import de.fernunihagen.rtt.framework.IElementCollector;
import de.fernunihagen.rtt.framework.IElementFilter;

/**
 * Java project collector.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class JavaProjectCollector implements
		IElementCollector<IJavaProject> {
	@Invar("null != $target")
	private final List<IJavaProject> _list;
	@Invar("null != $target")
	private final IElementFilter<IAdaptable> _filter;
	@Invar("null != $target")
	private final IProject[] _scope;

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				JavaProjectCollector.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Create a java project collector, used to fill given list with matching
	 * elements. List has to be synchronized in case collector is used
	 * concurrently.
	 * 
	 * @param list
	 *            List
	 * @param filter
	 *            Filter
	 * @param scope
	 *            Projects in search scope
	 */
	@Pre("null != list && null != filter && null != scope")
	public JavaProjectCollector(final List<IJavaProject> list,
			final IElementFilter<IAdaptable> filter, final IProject[] scope) {
		_list = list;
		_filter = filter;
		_scope = scope;
	}

	/**
	 * Creates a java project collector, used to collect matching elements.
	 * 
	 * @param filter
	 *            Filter
	 * @param scope
	 *            Projects in search scope
	 */
	@Pre("null != filter && null != scope")
	public JavaProjectCollector(final IElementFilter<IAdaptable> filter,
			final IProject[] scope) {
		this(new ArrayList<IJavaProject>(scope.length), filter, scope);
	}

	/**
	 * Creates a java project collector.
	 * 
	 * @param scope
	 *            Projects in search scope
	 */
	@Pre("null != scope")
	public JavaProjectCollector(final IProject[] scope) {
		this(IElementFilter.ALL, scope);
	}

	@Override
	public void collect(final IProgressMonitor pm) throws CoreException {
		try {
			clear();

			final IProject[] projects = getScope();
			final SubMonitor monitor = SubMonitor.convert(pm,
					"Collecting Java project(s)", projects.length);
			getLogger().debug("Collecting Java project(s)");

			for (IProject project : projects) {
				monitor.subTask(project.getName());
				getLogger().debug(project.getName());

				project.open(monitor.newChild(1));

				final boolean fIsJavaProject = project
						.hasNature(JavaCore.NATURE_ID);

				if (fIsJavaProject) {
					final IJavaProject javaProject = JavaCore.create(project);
					final boolean fAccept = _filter.accept(javaProject);

					if (getLogger().isDebugEnabled()) {
						final String message = String.format("[P/ %s (%s)]",
								javaProject.getElementName(), fAccept);
						getLogger().debug(message);
					}

					if (fAccept) {
						_list.add(javaProject);
					}
				}
			}

			if (getLogger().isDebugEnabled()) {
				final String message = String.format("Collection done (%d)",
						_list.size());
				getLogger().debug(message);
			}

		} finally {
			if (null != pm) {
				pm.done();
			}
		}

	}

	@Override
	@Post
	public List<? extends IJavaProject> getMatches() {
		return Collections.unmodifiableList(_list);
	}

	@Override
	@Post
	public IProject[] getScope() {
		return _scope;
	}

	@Override
	public void clear() {
		_list.clear();
	}

}
