package de.fernunihagen.rtt.internal.jdt.oracle;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;
import static de.fernunihagen.rtt.internal.util.ClassUtil.getClassUtil;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;
import org.eclipse.core.runtime.jobs.IJobManager;
import org.eclipse.core.runtime.jobs.Job;

import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IType;

import org.eclipse.jdt.junit.model.ITestCaseElement;
import org.eclipse.jdt.junit.model.ITestElement;
import org.eclipse.jdt.junit.model.ITestSuiteElement;

import de.fernunihagen.rtt.framework.ElementFilterOp;
import de.fernunihagen.rtt.framework.IElementCollector;
import de.fernunihagen.rtt.framework.IElementFilter;
import de.fernunihagen.rtt.framework.ITestElementContainer;
import de.fernunihagen.rtt.framework.ITestContext;
import de.fernunihagen.rtt.framework.ITestRunSession;

import de.fernunihagen.rtt.framework.ITestElement.Result;
import de.fernunihagen.rtt.framework.ITestElement.ProgressState;

import de.fernunihagen.rtt.framework.oracle.Oracle;
import de.fernunihagen.rtt.framework.oracle.OracleRunDescription;

import de.fernunihagen.rtt.internal.jdt.ExcludedJavaElementFilter;
import de.fernunihagen.rtt.internal.jdt.IJUnitTestRunListener;
import de.fernunihagen.rtt.internal.jdt.JUnit4TestRun;
import de.fernunihagen.rtt.internal.jdt.JUnitTestCaseCollector;
import de.fernunihagen.rtt.internal.jdt.JUnitTestRun;
import de.fernunihagen.rtt.internal.jdt.JUnitTestRunNotification;
import de.fernunihagen.rtt.internal.jdt.JavaTestElement;

/**
 * JUnit oracle, semantic oracle used to verify that test targets satisfies
 * included JUnit test cases.
 * 
 * @author Osama El Hosami
 */
public final class JUnitOracle extends Oracle implements
		IJUnitTestRunListener<JUnitOracleRunState> {

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				JUnitOracle.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Creates a JUnit oracle
	 */
	public JUnitOracle() {
		super();
	}

	@Pre
	@Post
	@Override
	protected JUnitOracleRunState getState(final ITestContext context) {
		return (JUnitOracleRunState) super.getState(context);
	}

	private final <T> T as(final IAdaptable adaptable,
			final Class<? extends T> clazz) {
		return getClassUtil().as(adaptable, clazz);
	}

	private void reportError(final CoreException[] exceptions,
			final String message) throws CoreException {
		getRTTCorePlugin().reportError(exceptions, message);
	}

	private void reportError(final String message, final Throwable e)
			throws CoreException {
		getRTTCorePlugin().reportError(message, e);
	}

	@Pre("null != testcases && null != testcaseHandles")
	private int getHandlesOfTestCasesInSameProject(
			final List<? extends IAdaptable> testcases,
			final List<String> testcaseHandles, int start) {

		IJavaProject project = null;
		int i;

		for (i = start; i < testcases.size(); i++) {
			IType testcase = as(testcases.get(i), IType.class);
			Assert.isTrue(null != testcase);

			if (null == project) {
				project = testcase.getJavaProject();
			}

			if (!project.equals(testcase.getJavaProject())) {
				break;
			}

			testcaseHandles.add(testcase.getHandleIdentifier());
		}

		return i;
	}

	private Result toResult(final ITestElement.Result result) {
		if (result == ITestElement.Result.OK) {
			return Result.OK;
		}
		if (result == ITestElement.Result.ERROR) {
			return Result.ERROR;
		}
		if (result == ITestElement.Result.FAILURE) {
			return Result.FAILURE;
		}
		if (result == ITestElement.Result.IGNORED) {
			return Result.IGNORED;
		}
		return Result.UNDEFINED;
	}

	@Override
	@Pre("null != o && null != arg")
	public void update(final JUnitTestRun<JUnitOracleRunState> o,
			final JUnitTestRunNotification arg) {
		final JUnitOracleRunState state = o.getSubject();

		ITestElementContainer parent = state
				.getParentContainer(arg.testelement);
		if (null == parent) {
			return;
		}

		switch (arg.type) {
		case SESSION_STARTED:
			break;
		case SESSION_FINISHED:
			parent.setTestResult(toResult(arg.testelement.getTestResult(false)));
			break;
		case TESTCASE_STARTED:
			if (arg.testelement instanceof ITestSuiteElement) {
				final JUnitTestSuiteElementContainer element = new JUnitTestSuiteElementContainer(
						parent, (ITestSuiteElement) arg.testelement);
				state.addContainer(element);
				element.setProgressState(ProgressState.RUNNING);
			} else {
				final JUnitTestCaseElement element = new JUnitTestCaseElement(
						parent, o.getProject(),
						(ITestCaseElement) arg.testelement);
				state.addElement(element);
				element.setProgressState(ProgressState.RUNNING);
			}
			break;
		case TESTCASE_FINISHED:
			if (arg.testelement instanceof ITestSuiteElement) {
				final ITestSuiteElement testsuite = (ITestSuiteElement) arg.testelement;
				JUnitTestSuiteElementContainer element = (JUnitTestSuiteElementContainer) state
						.getContainer(testsuite);
				Assert.isNotNull(element);
				element.setTestResult(toResult(testsuite.getTestResult(false)));
				state.removeContainer(testsuite);
			} else {
				final ITestCaseElement testcase = (ITestCaseElement) arg.testelement;
				JUnitTestCaseElement element = state.getElement(testcase);
				Assert.isNotNull(element);
				element.setTestResult(toResult(testcase.getTestResult(false)),
						testcase.getFailureTrace());
				state.removeElement(testcase);
			}
			break;
		}
	}

	@Post
	@Pre
	@Override
	public IElementCollector<? extends IAdaptable> getCollector(
			final IElementFilter<IAdaptable> filter, final IProject[] scope) {
		return new JUnitTestCaseCollector(filter, scope);
	}

	@Override
	@Pre
	@Post
	public JUnitOracleRunState create(final ITestRunSession session,
			final IProgressMonitor monitor) throws CoreException {

		final JUnitOracleRunState state;

		try {
			final IProject[] scope = session.getScope();
			final OracleRunDescription description = new OracleRunDescription(
					session.getConfiguration(this));
			IElementFilter<IAdaptable> filter = new ExcludedJavaElementFilter(
					description.getExcludedElementIdentifiers());
			IElementCollector<? extends IAdaptable> collector = getCollector(
					filter, scope);
			state = new JUnitOracleRunState(this, collector, description);
			state.updateTargets(monitor);

			// FIXME: Excluded tests are used as targets, but while a indication
			// of included test suites using those tests is missing, this
			// behavior is undesired. As workaround we set all tests as indirect
			// targets.

			// Research for all tests included and excluded
			collector = getCollector(IElementFilter.ALL, scope);
			collector.collect(null);

			session.addIndirectTargetFilter(ElementFilterOp
					.adaptToResourceFilter(collector));
		} finally {
			if (null != monitor) {
				monitor.done();
			}
		}

		return state;
	}

	@Override
	@Post
	public Type getType() {
		return Type.SEMANTIC;
	}

	@Override
	@Pre
	public void start(final ITestContext context, final IProgressMonitor pm)
			throws CoreException {
		try {
			int offset = 0;
			final JUnitOracleRunState state = getState(context);
			final List<? extends IAdaptable> targets = state.getTargets();
			final int max = targets.size();
			final List<String> testcaseHandles = new ArrayList<String>(max);
			final List<CoreException> errors = new ArrayList<CoreException>();
			final boolean fShouldContinueOnError = shouldContinueOnError(context);
			final List<JUnitTestRun<JUnitOracleRunState>> testruns = state
					.getTestRuns();
			final ITestElementContainer container = context
					.createChildContainer(max, "Launches");

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Creating launch(es) for junit test targets", max);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format(
						"Creating launches for junit test target(s) (%s)", max);
				getLogger().debug(message);
			}

			Assert.isTrue(testruns.size() == 0);
			container.setProgressState(ProgressState.RUNNING);

			while (offset < max) {
				if (monitor.isCanceled()) {
					throw new OperationCanceledException();
				}

				final IJavaProject project = as(targets.get(offset),
						IType.class).getJavaProject();

				final String testName = project.getElementName();
				monitor.subTask(testName);

				final JavaTestElement element = new JavaTestElement(container,
						project);

				element.setProgressState(ProgressState.RUNNING);

				if (errors.isEmpty() || fShouldContinueOnError) {
					element.setTestResult(Result.OK);

					offset = getHandlesOfTestCasesInSameProject(targets,
							testcaseHandles, offset);

					try {
						// FIXME: make test run type configurable
						final JUnitTestRun<JUnitOracleRunState> testrun = new JUnit4TestRun<JUnitOracleRunState>(
								testName, project, testcaseHandles, state);
						testrun.addObserver(this);
						testruns.add(testrun);

					} catch (OperationCanceledException e) {
						element.setProgressState(ProgressState.STOPPED);
						throw e;
					} catch (CoreException e) {
						getLogger().trace("Unexpected", e);
						element.setTestResult(Result.ERROR);
						errors.add(e);
					}
				}
				// else ignored notifications

				element.setProgressState(ProgressState.COMPLETED);

				if (getLogger().isDebugEnabled()) {
					final String message = String.format("%s (%s;%s;%.3fs)",
							element, element.getTestResult(),
							element.getProgressState(true),
							element.getElapsedTimeInSeconds());
					getLogger().debug(message);
				}

				monitor.worked(offset);
				testcaseHandles.clear();
			}

			getLogger().assertLog(offset == max,
					"Post-condition failed - some test cases may be missing");

			container.setProgressState(ProgressState.COMPLETED);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format(
						"Creation done (%s;%s;%.3fs)",
						container.getTestResult(),
						container.getProgressState(true),
						container.getElapsedTimeInSeconds());
				getLogger().debug(message);
			}

			if (!errors.isEmpty()) {
				reportError(errors.toArray(new CoreException[errors.size()]),
						"Error ocurred while creating launch configuration(s)");
			}

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	private void joinBuild(final IProgressMonitor pm) throws CoreException {
		try {
			final IJobManager jobManager = Job.getJobManager();
			final SubMonitor monitor = SubMonitor.convert(pm,
					"Waiting for build to complete", 2);

			boolean wait = jobManager.find(ResourcesPlugin.FAMILY_AUTO_BUILD).length > 0
					|| jobManager.find(ResourcesPlugin.FAMILY_MANUAL_BUILD).length > 0;
			if (wait) {
				try {
					jobManager.join(ResourcesPlugin.FAMILY_MANUAL_BUILD,
							monitor.newChild(1));
					jobManager.join(ResourcesPlugin.FAMILY_AUTO_BUILD,
							monitor.newChild(1));
				} catch (InterruptedException e) {
					reportError(null, e);
				}
			}
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Override
	@Pre
	public void apply(final ITestContext context, final IProgressMonitor pm)
			throws CoreException {

		final JUnitOracleRunState state = getState(context);

		try {
			final List<CoreException> errors = new ArrayList<CoreException>();
			final boolean fShouldContinueOnError = shouldContinueOnError(context);
			final List<JUnitTestRun<JUnitOracleRunState>> testruns = state
					.getTestRuns();
			boolean fStatisfied = context.getTestResult().isSuccess();
			final ITestElementContainer container = context
					.createChildContainer(testruns.size(), "Verifications");

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Launching junit test(s)", testruns.size() + 10);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format(
						"Launching junit test(s) (%s)", testruns.size());
				getLogger().debug(message);
			}

			joinBuild(monitor.newChild(10));

			container.setProgressState(ProgressState.RUNNING);

			for (JUnitTestRun<JUnitOracleRunState> testrun : testruns) {
				monitor.subTask(testrun.toString());

				final ITestElementContainer element = new JUnitTestSessionElementContainer(
						container, testrun);
				state.addContainer(element);

				if (fStatisfied || fShouldContinueOnError) {
					try {
						testrun.start(monitor.newChild(1));

						fStatisfied = ITestElement.Result.OK == testrun
								.getTestResult();

					} catch (OperationCanceledException e) {
						element.setProgressState(ProgressState.STOPPED);
						throw e;
					} catch (CoreException e) {
						getLogger().trace("Unexpected", e);
						element.setTestResult(Result.ERROR);
						fStatisfied = false;
						errors.add(e);
					}
				}
				// else ignored notifications

				state.removeContainer(element.getLabel());
			}

			container.setProgressState(ProgressState.COMPLETED);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format(
						"Launching done (%s;%s;%.3fs)",
						container.getTestResult(),
						container.getProgressState(true),
						container.getElapsedTimeInSeconds());
				getLogger().debug(message);
			}

			if (!errors.isEmpty()) {
				reportError(errors.toArray(new CoreException[errors.size()]),
						"Error ocurred while launching JUnit test(s)");
			}

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}
}
