package de.fernunihagen.rtt.internal.jdt.oracle;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceDescription;
import org.eclipse.core.resources.IncrementalProjectBuilder;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IAdaptable;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IProblemRequestor;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.WorkingCopyOwner;

import org.eclipse.jdt.core.compiler.IProblem;

import de.fernunihagen.rtt.framework.ITestElement.ProgressState;
import de.fernunihagen.rtt.framework.ITestElement.Result;

import de.fernunihagen.rtt.framework.IElementCollector;
import de.fernunihagen.rtt.framework.IElementFilter;
import de.fernunihagen.rtt.framework.ITestContext;
import de.fernunihagen.rtt.framework.ITestElementContainer;
import de.fernunihagen.rtt.framework.ITestRunSession;

import de.fernunihagen.rtt.framework.oracle.Oracle;
import de.fernunihagen.rtt.framework.oracle.OracleRunState;
import de.fernunihagen.rtt.framework.oracle.OracleRunDescription;

import de.fernunihagen.rtt.internal.jdt.CompilationUnitCollector;
import de.fernunihagen.rtt.internal.jdt.ExcludedJavaElementFilter;
import de.fernunihagen.rtt.internal.jdt.JavaCompileProblemTestElement;
import de.fernunihagen.rtt.internal.jdt.JavaTestElement;
import de.fernunihagen.rtt.internal.jdt.JavaTestElementContainer;

import de.fernunihagen.rtt.internal.util.MutableBoolean;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import static de.fernunihagen.rtt.internal.util.ClassUtil.getClassUtil;

/**
 * Java compile oracle, syntactic oracle used to test that test targets are
 * compilable.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class JavaCompileOracle extends Oracle {

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				JavaCompileOracle.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Creates a compile oracle.
	 */
	public JavaCompileOracle() {
		super();
	}

	@Override
	@Post
	public Type getType() {
		return Type.SYNTACTIC;
	}

	private IWorkspace getWorkspace() {
		return getRTTCorePlugin().getWorkspace();
	}

	private IWorkspaceDescription getWorkspaceDescription() {
		return getWorkspace().getDescription();
	}

	private final <T> T as(final IAdaptable adaptable,
			final Class<? extends T> clazz) {
		return getClassUtil().as(adaptable, clazz);
	}

	private void reportError(final CoreException[] exceptions,
			final String message) throws CoreException {
		getRTTCorePlugin().reportError(exceptions, message);
	}

	@Pre("null != context && null != parent && null != monitor")
	private void build(final ITestContext context,
			final ITestElementContainer parent, final SubMonitor monitor)
			throws CoreException {

		final IProject[] targets = context.getScope();
		final List<CoreException> errors = new ArrayList<CoreException>();
		final boolean fShouldContinueOnError = shouldContinueOnError(context);
		final ITestElementContainer container = parent.createChildContainer(
				targets.length, "Builds");

		monitor.beginTask("Building project(s)", targets.length);
		getLogger().debug("Building project(s)");

		container.setProgressState(ProgressState.RUNNING);

		for (IProject target : targets) {
			final JavaTestElement element = new JavaTestElement(container,
					JavaCore.create(target));

			element.setProgressState(ProgressState.RUNNING);

			if (errors.isEmpty() || fShouldContinueOnError) {
				element.setTestResult(Result.OK);

				try {
					target.build(IncrementalProjectBuilder.INCREMENTAL_BUILD,
							monitor.newChild(1));
				} catch (OperationCanceledException e) {
					element.setProgressState(ProgressState.STOPPED);
					throw e;
				} catch (CoreException e) {
					getLogger().trace("Unexpected", e);
					element.setTestResult(Result.ERROR);
					errors.add(e);
				}
			}
			// else ignored notifications

			element.setProgressState(ProgressState.COMPLETED);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format("%s (%s;%s;%.3fs)",
						element, element.getTestResult(), element
								.getProgressState(true), element
								.getElapsedTimeInSeconds());
				getLogger().debug(message);
			}
		}

		container.setProgressState(ProgressState.COMPLETED);

		if (getLogger().isDebugEnabled()) {
			final String message = String.format("Build done (%s;%s;%.3fs)",
					container.getTestResult(),
					container.getProgressState(true), container
							.getElapsedTimeInSeconds());
			getLogger().debug(message);
		}

		if (!errors.isEmpty()) {
			reportError(errors.toArray(new CoreException[errors.size()]),
					"Error ocurred while cleaning project(s)");
		}
	}

	@Override
	@Pre("null != context")
	public void start(final ITestContext context, final IProgressMonitor pm)
			throws CoreException {

		try {
			final IProject[] targets = context.getScope();
			final ITestElementContainer container = context
					.createChildContainer(targets.length, "Cleans");
			final List<CoreException> errors = new ArrayList<CoreException>();
			final boolean fShouldContinueOnError = shouldContinueOnError(context);

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Cleaning project(s)", targets.length * 2);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format(
						"Cleaning project(s) (%s)", targets.length);
				getLogger().debug(message);
			}

			container.setProgressState(ProgressState.RUNNING);

			for (IProject target : targets) {
				final JavaTestElement element = new JavaTestElement(container,
						JavaCore.create(target));

				element.setProgressState(ProgressState.RUNNING);

				if (errors.isEmpty() || fShouldContinueOnError) {
					element.setTestResult(Result.OK);

					try {
						target.open(monitor.newChild(1));
						target.build(IncrementalProjectBuilder.CLEAN_BUILD,
								monitor.newChild(1));
					} catch (OperationCanceledException e) {
						element.setProgressState(ProgressState.STOPPED);
						throw e;
					} catch (CoreException e) {
						getLogger().trace("Unexpected", e);
						element.setTestResult(Result.ERROR);
						errors.add(e);
					}
				}
				// else ignored notifications

				element.setProgressState(ProgressState.COMPLETED);

				if (getLogger().isDebugEnabled()) {
					final String message = String.format("%s (%s;%s;%.3fs)",
							element, element.getTestResult(), element
									.getProgressState(true), element
									.getElapsedTimeInSeconds());
					getLogger().debug(message);
				}
			}

			container.setProgressState(ProgressState.COMPLETED);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format(
						"Clean done (%s;%s;%.3fs)", container.getTestResult(),
						container.getProgressState(true), container
								.getElapsedTimeInSeconds());
				getLogger().debug(message);
			}

			if (!errors.isEmpty()) {
				reportError(errors.toArray(new CoreException[errors.size()]),
						"Error ocurred while cleaning project(s)");
			}

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Pre("null != testElementRef && null != compilable")
	private WorkingCopyOwner getWorkingCopyOwner(
			final ITestElementContainer[] testElementRef,
			final MutableBoolean compilable,
			final boolean fShouldContinueOnError) {

		final IProblemRequestor requestor = new IProblemRequestor() {
			@Override
			@Pre("null != problem")
			public void acceptProblem(final IProblem problem) {
				if (problem.isError()) {
					Assert.isNotNull(testElementRef[0]);
					new JavaCompileProblemTestElement(testElementRef[0],
							problem);

					compilable.setValue(false);
				}
			}

			@Override
			public void beginReporting() {
			}

			@Override
			public void endReporting() {
			}

			@Override
			public boolean isActive() {
				return compilable.getValue() || fShouldContinueOnError;
			}

		};

		return new WorkingCopyOwner() {
			@Override
			public IProblemRequestor getProblemRequestor(
					final ICompilationUnit unit) {
				return requestor;
			}
		};
	}

	@Override
	@Pre("null != context")
	public void apply(final ITestContext context, final IProgressMonitor pm)
			throws CoreException {

		try {
			final MutableBoolean compilable = new MutableBoolean(context
					.getTestResult().isSuccess());
			final ITestElementContainer[] testElementRef = { null };
			final List<CoreException> errors = new ArrayList<CoreException>();
			final boolean fShouldContinueOnError = shouldContinueOnError(context);
			final List<? extends IAdaptable> targets = getState(context)
					.getTargets();
			final WorkingCopyOwner owner = getWorkingCopyOwner(testElementRef,
					compilable, fShouldContinueOnError);
			final ITestElementContainer container = context
					.createChildContainer(targets.size(), "Verifications");

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Verifying that test targets are compilable", targets
							.size() + 100);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format(
						"Verifying that test targets (%s) are compilable",
						targets.size());
				getLogger().debug(message);
			}

			container.setProgressState(ProgressState.RUNNING);

			for (IAdaptable target : targets) {
				final ICompilationUnit unit = as(target, ICompilationUnit.class);

				Assert.isTrue(null != unit);
				monitor.subTask(unit.getElementName());

				final ITestElementContainer element = new JavaTestElementContainer(
						container, unit);

				element.setProgressState(ProgressState.RUNNING);

				if (compilable.getValue() || fShouldContinueOnError) {
					testElementRef[0] = element;

					try {
						unit.getWorkingCopy(owner, monitor.newChild(1));
						unit.discardWorkingCopy();
					} catch (OperationCanceledException e) {
						element.setProgressState(ProgressState.STOPPED);
						throw e;
					} catch (CoreException e) {
						getLogger().trace("Unexpected", e);
						compilable.setValue(false);
						element.setTestResult(Result.ERROR);
						errors.add(e);
					}

					testElementRef[0] = null;
				}
				// else ignored notifications

				// for performance reason we add only failed test elements
				if (!element.getTestResult().isSuccess()) {
					container.add(element);
				}

				element.setProgressState(ProgressState.COMPLETED);

				if (getLogger().isDebugEnabled()) {
					final String message = String.format("%s (%s;%s;%.3fs)",
							element, element.getTestResult(), element
									.getProgressState(true), element
									.getElapsedTimeInSeconds());
					getLogger().debug(message);
				}
			}

			// build projects if auto build is disabled
			if (compilable.getValue()
					&& !getWorkspaceDescription().isAutoBuilding()) {
				try {
					build(context, container, monitor.newChild(100));
				} catch (OperationCanceledException e) {
					container.setProgressState(ProgressState.STOPPED);
					throw e;
				} catch (CoreException e) {
					getLogger().trace("Unexpected", e);
					container.setTestResult(Result.ERROR);
					errors.add(e);
				}
			}

			if (compilable.getValue()) {
				container.setTestResult(Result.OK);
			}

			container.setProgressState(ProgressState.COMPLETED);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format(
						"Verification done (%s;%s;%.3fs)", container
								.getTestResult(), container
								.getProgressState(true), container
								.getElapsedTimeInSeconds());
				getLogger().debug(message);
			}

			if (!errors.isEmpty()) {
				reportError(errors.toArray(new CoreException[errors.size()]),
						"Error ocurred while verifying that test target(s) are compilable");
			}
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Post
	@Pre
	@Override
	public IElementCollector<? extends IAdaptable> getCollector(
			final IElementFilter<IAdaptable> filter, final IProject[] scope) {
		return new CompilationUnitCollector(filter, scope);
	}

	@Override
	@Pre
	@Post
	public OracleRunState create(final ITestRunSession session,
			final IProgressMonitor monitor) throws CoreException {

		final OracleRunState state;

		// disable auto build to force oracle building projects
//		final IWorkspaceDescription wkDescription = getWorkspaceDescription();
//		wkDescription.setAutoBuilding(false);
//		getWorkspace().setDescription(wkDescription);

		try {
			final IProject[] scope = session.getScope();
			final OracleRunDescription description = new OracleRunDescription(
					session.getConfiguration(this));
			final IElementFilter<IAdaptable> filter = new ExcludedJavaElementFilter(
					description.getExcludedElementIdentifiers());
			state = new OracleRunState(this, getCollector(filter, scope),
					description);
			state.updateTargets(monitor);
		} finally {
			if (null != monitor) {
				monitor.done();
			}
		}

		return state;
	}
}
