package de.fernunihagen.rtt.internal.runtime;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import static de.fernunihagen.rtt.framework.RTTCore.getTestRunListenerRegistry;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubMonitor;

import de.fernunihagen.rtt.framework.ITestApplication;
import de.fernunihagen.rtt.framework.ITestRunListener;

import de.fernunihagen.rtt.framework.model.TestRootRunSession;

/**
 * Abstract test application.
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class AbstractTestApplication extends TestRootRunSession
		implements ITestApplication {

	private boolean _fImportProjects;

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				AbstractTestApplication.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Creates a test application.
	 */
	public AbstractTestApplication() {
		super();
		setRequestProjects(true);
	}

	/**
	 * Parses the program arguments
	 * 
	 * @param args
	 *            Arguments
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	@Pre("null != args")
	protected abstract void parseArguments(final String[] args)
			throws CoreException;

	/**
	 * Imports test projects
	 * 
	 * @param monitor
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	protected abstract void importProjects(final IProgressMonitor monitor)
			throws CoreException;

	/**
	 * Runs the test application
	 * 
	 * @param monitor
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	protected abstract void runImpl(final IProgressMonitor monitor)
			throws CoreException;

	/**
	 * Sets the state of request to import projects.
	 * 
	 * @param fState
	 *            Flag indicate request state
	 */
	protected void setRequestProjects(final boolean fState) {
		_fImportProjects = fState;
	}

	/**
	 * Runs the test application
	 * 
	 * @param pm
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	public final void run(final IProgressMonitor pm) throws CoreException {
		try {
			final SubMonitor monitor = SubMonitor.convert(pm,
					"Running tool test(s)", 2);

			setProgressState(ProgressState.RUNNING);

			if (_fImportProjects) {
				importProjects(monitor.newChild(1));
				setRequestProjects(false);
			}

			monitor.setWorkRemaining(1);
			runImpl(monitor.newChild(1));

			setProgressState(ProgressState.COMPLETED);
		} catch (CoreException e) {
			setTestResult(Result.ERROR);
			setProgressState(ProgressState.COMPLETED);
			throw e;
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	private void registerObservers() throws CoreException {
		deleteObservers();
		final ITestRunListener[] listeners = getTestRunListenerRegistry().get();
		for (ITestRunListener listener : listeners) {
			addObserver(listener);
		}
	}

	@Override
	@Pre
	@Post
	public IStatus run(final String[] args, final IProgressMonitor pm) {
		IStatus status = Status.OK_STATUS;

		try {
			registerObservers();
			parseArguments(args);
			run(pm);

		} catch (OperationCanceledException e) {
			setProgressState(ProgressState.STOPPED);
			status = Status.CANCEL_STATUS;
		} catch (CoreException e) {
			setTestResult(Result.ERROR);
			setProgressState(ProgressState.COMPLETED);
			getLogger().trace("Unexpected", e);
			status = e.getStatus();
		} finally {
			if (null != pm) {
				pm.done();
			}
		}

		return status;
	}
}
