package de.fernunihagen.rtt.internal.runtime;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import static de.fernunihagen.rtt.framework.RTTCore.getOracleProvider;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;

import de.fernunihagen.rtt.framework.ITestContext;
import de.fernunihagen.rtt.framework.ITestElement;
import de.fernunihagen.rtt.framework.ITestElementContainer;
import de.fernunihagen.rtt.framework.ITestRunSession;

import de.fernunihagen.rtt.framework.ITestElement.ProgressState;
import de.fernunihagen.rtt.framework.ITestElement.Result;

import de.fernunihagen.rtt.framework.model.ExceptionTestElement;
import de.fernunihagen.rtt.framework.model.TestContext;

import de.fernunihagen.rtt.framework.oracle.IOracle;
import de.fernunihagen.rtt.framework.oracle.IOracleRunner;

import de.fernunihagen.rtt.internal.util.ArrayIterator;

/**
 * Oracle runner.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class OracleRunner extends TestRunner implements IOracleRunner {
	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				OracleRunner.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Creates an oracle runner.
	 */
	public OracleRunner() {
		super();
	}

	private Map<String, IOracle> getOracles() throws CoreException {
		return getOracleProvider().getMap();
	}

	private void reportError(final CoreException[] exceptions,
			final String message) throws CoreException {
		getRTTCorePlugin().reportError(exceptions, message);
	}

	@Pre
	@Post
	@Override
	protected OracleRunnerRunState getState(final ITestContext context) {
		return (OracleRunnerRunState) super.getState(context);
	}

	@Pre
	@Override
	public void apply(final ITestContext context, final IProgressMonitor pm)
			throws CoreException {

		try {
			// FIXME make configurable
			final boolean fShouldContinueOnError = false;
			boolean fSeekNext = true;
			IOracle oracle = null;
			boolean fSatisfied = context.getTestResult().isSuccess();
			final OracleRunnerRunState state = getState(context);
			final IOracle[] oracles = state.getActors();
			int workRemaining = oracles.length;
			final List<CoreException> errors = new ArrayList<CoreException>();
			final Iterator<IOracle> ita = new ArrayIterator<IOracle>(oracles);
			final ITestElementContainer container = context;

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Running oracle(s)", 2 * workRemaining);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format("Running oracle(s) (%s)",
						workRemaining);
				getLogger().debug(message);
			}

			container.setProgressState(ProgressState.RUNNING);

			while (ita.hasNext()) {
				if (fSeekNext) {
					oracle = ita.next();
					workRemaining--;
				}

				monitor.subTask(oracle.getName());
				ITestElement element;

				if (fSatisfied || !fSeekNext || fShouldContinueOnError) {
					TestContext testContext = null;
					try {
						element = testContext = state.createContext(oracle,
								container, monitor.newChild(1));
					} catch (CoreException e) {
						getLogger().trace("Unexpected", e);
						element = new ExceptionTestElement(container, oracle, e);
						fSeekNext = true;
						fSatisfied = false;
						errors.add(e);
					}

					if (null != testContext) {
						testContext.setProgressState(ProgressState.RUNNING);

						try {
							fSeekNext = testContext.run(monitor.newChild(1));
							fSatisfied = container.getTestResult().isSuccess();
						} catch (OperationCanceledException e) {
							testContext.setProgressState(ProgressState.STOPPED);
							throw e;
						} catch (CoreException e) {
							getLogger().trace("Unexpected", e);
							fSeekNext = true;
							fSatisfied = false;
							testContext.setTestResult(Result.ERROR);
							errors.add(e);
						}

						testContext.setProgressState(ProgressState.COMPLETED);
					}

				} else {
					// ignored notifications
					element = container.createChild(oracle.getName(),
							Result.IGNORED);
				}

				if (getLogger().isDebugEnabled()) {
					final String message = String.format("%s (%s;%s;%.3fs)",
							element, element.getTestResult(), element
									.getProgressState(true), element
									.getElapsedTimeInSeconds());
					getLogger().debug(message);
				}

				monitor.setWorkRemaining(2 * workRemaining);
			}

			container.setProgressState(ProgressState.COMPLETED);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format("Run done (%s;%s;%.3fs)",
						container.getTestResult(), container
								.getProgressState(true), container
								.getElapsedTimeInSeconds());
				getLogger().debug(message);
			}

			if (!errors.isEmpty()) {
				reportError(errors.toArray(new CoreException[errors.size()]),
						"Failed to run oracle(s), least one actor failed to run");
			}
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Override
	@Post
	@Pre
	public OracleRunnerRunState create(final ITestRunSession session,
			final IProgressMonitor monitor) throws CoreException {
		try {
			return new OracleRunnerRunState(this, session, getOracles());
		} finally {
			if (null != monitor) {
				monitor.done();
			}
		}
	}
}
