package de.fernunihagen.rtt.internal.runtime;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import static de.fernunihagen.rtt.framework.RTTCore.getRefactoringTestAdapterProvider;
import static de.fernunihagen.rtt.framework.RTTCore.getOracleRunner;

import java.util.Map;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;

import de.fernunihagen.rtt.framework.ITestActorState;
import de.fernunihagen.rtt.framework.ITestContext;
import de.fernunihagen.rtt.framework.ITestElement;
import de.fernunihagen.rtt.framework.ITestElementContainer;
import de.fernunihagen.rtt.framework.ITestRunSession;

import de.fernunihagen.rtt.framework.ITestElement.ProgressState;
import de.fernunihagen.rtt.framework.ITestElement.Result;

import de.fernunihagen.rtt.framework.model.ExceptionTestElement;
import de.fernunihagen.rtt.framework.model.TestContext;

import de.fernunihagen.rtt.framework.testadapter.IRefactoringTestAdapter;
import de.fernunihagen.rtt.framework.testadapter.IRefactoringTestRunner;
import de.fernunihagen.rtt.framework.testadapter.RefactoringTestRunState;

/**
 * Refactoring test runner.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class RefactoringTestRunner extends TestRunner implements
		IRefactoringTestRunner {

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				RefactoringTestRunner.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Creates a refactoring runner.
	 */
	public RefactoringTestRunner() {
		super();
	}

	private Map<String, IRefactoringTestAdapter> getRefactoringTestAdapters()
			throws CoreException {
		return getRefactoringTestAdapterProvider().getMap();
	}

	@Pre
	@Post
	@Override
	protected RefactoringTestRunnerRunState getState(final ITestContext context) {
		return (RefactoringTestRunnerRunState) super.getState(context);
	}

	@Override
	@Pre
	@Post
	public RefactoringTestRunnerRunState create(final ITestRunSession session,
			final IProgressMonitor monitor) throws CoreException {
		try {
			return new RefactoringTestRunnerRunState(this, session,
					getRefactoringTestAdapters(), getOracleRunner());
		} finally {
			if (null != monitor) {
				monitor.done();
			}
		}
	}

	@Override
	@Pre
	public boolean canUndoChange(final ITestContext context) {
		try {
			final RefactoringTestRunnerRunState state = getState(context);
			final ITestContext testContext = state.getUndoContext();

			return null != testContext
					&& state.getTestAdapter().canUndoChange(testContext);

		} catch (CoreException e) {
			getLogger().trace("Unexpected", e);
		}
		return false;
	}

	@Override
	@Pre
	public void undoChange(final ITestContext context, final IProgressMonitor pm)
			throws CoreException {
		try {
			final RefactoringTestRunnerRunState state = getState(context);

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Undoing change(s)", 1);

			getLogger().debug("Undoing change(s)");

			if (canUndoChange(context)) {
				final IRefactoringTestAdapter testadapter = state
						.getTestAdapter();
				testadapter.undoChange(state.getUndoContext(), monitor
						.newChild(1));
				state.setUndoContext(null);
			} else {
				context.createChild("Undo", context.getTestResult());
			}

			getLogger().debug("Undo done");

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Pre("null != state && null != container")
	@Post("null != $return")
	private TestContext executeOracleRunner(
			final RefactoringTestRunnerRunState state,
			final ITestElementContainer container, final IProgressMonitor pm)
			throws CoreException {

		try {

			final SubMonitor monitor = SubMonitor.convert(pm);

			container.setProgressState(ProgressState.RUNNING);

			final TestContext oracleRunnerContext = state
					.createOracleRunnerContext(container, monitor.newChild(1));

			while (!oracleRunnerContext.run(monitor.newChild(1))) {

			}

			container.setProgressState(ProgressState.COMPLETED);

			return oracleRunnerContext;

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Pre("null != state && null != context")
	private boolean preConditionCheck(
			final RefactoringTestRunnerRunState state,
			final ITestElementContainer container,
			final boolean fInitialResult, final IProgressMonitor monitor)
			throws CoreException {

		final ITestElementContainer childContainer = container
				.createChildContainer(0, "Pre-Condition-Check");

		if (!fInitialResult) {
			final TestContext oracleRunnerContext = executeOracleRunner(state,
					childContainer, monitor);

			return oracleRunnerContext.getTestResult().isSuccess();
		}

		childContainer.setProgressState(ProgressState.COMPLETED);

		return fInitialResult;
	}

	@Pre("null != state && null != container")
	private void postConditionCheck(final RefactoringTestRunnerRunState state,
			final ITestElementContainer container,
			final IProgressMonitor monitor) throws CoreException {

		final ITestElement[] children = container.getChildren();
		final ITestElement refactoringTestRunContext = children[children.length - 1];

		final ITestElementContainer childContainer = container
				.createChildContainer(0, "Post-Condition-Check");

		if (refactoringTestRunContext.getTestResult() == Result.OK) {
			executeOracleRunner(state, childContainer, monitor);
		} else {
			childContainer.setProgressState(ProgressState.COMPLETED);
		}
	}

	@Pre("null != state && null != container && null != testadapter && null != elementRef")
	@Post("null != elementRef[0]")
	private TestContext createContext(
			final RefactoringTestRunnerRunState state,
			final ITestElementContainer container,
			final IRefactoringTestAdapter testadapter,
			final ITestElement[] elementRef, final IProgressMonitor monitor) {
		TestContext testContext = null;
		try {
			testContext = state.createContext(testadapter, container, monitor);
			elementRef[0] = testContext;
		} catch (CoreException e) {
			getLogger().trace("Unexpected", e);
			elementRef[0] = new ExceptionTestElement(container, testadapter, e);
			state.setStalled(false);
			state.addError(e);
		}
		return testContext;
	}

	@Override
	@Pre
	public void apply(final ITestContext context, final IProgressMonitor pm)
			throws CoreException {
		try {
			final RefactoringTestRunnerRunState state = getState(context);

			if (state.hasNext()) {
				boolean fInitialResult = context.getTestResult().isSuccess();
				final IRefactoringTestAdapter testadapter = state
						.nextTestAdapter();
				final ITestElement[] elementRef = { null };
				final ITestElementContainer container = context;
				context.setTestResult(Result.NONE);

				final SubMonitor monitor = SubMonitor.convert(pm,
						"Running test adapter", 4);

				monitor.subTask(testadapter.getName());

				getLogger().debug("Running test adapter");

				final TestContext testContext = createContext(state, container,
						testadapter, elementRef, monitor.newChild(1));

				if (null != testContext) {
					state.setUndoContext(testContext);

					testContext.setProgressState(ProgressState.RUNNING);

					try {
						if (preConditionCheck(state, testContext,
								fInitialResult, monitor.newChild(1))) {

							state.setStalled(!testContext.run(monitor
									.newChild(1)));

							postConditionCheck(state, testContext, monitor
									.newChild(1));

							final boolean fSucceeded = testContext
									.getTestResult().isSuccess();
							final ITestActorState refactoringTestRunState = testContext
									.getState();

							// optimization: skip actual refactoring test
							// adapter if context indicates failure
							if (!fSucceeded
									&& refactoringTestRunState instanceof RefactoringTestRunState
									&& ((RefactoringTestRunState) refactoringTestRunState)
											.hasNext()) {

								final RefactoringTestRunState runState = (RefactoringTestRunState) refactoringTestRunState;

								runState.setStalled(false);
								runState.seekNext();

								state.setStalled(runState.hasNext());
							}
						}
					} catch (OperationCanceledException e) {
						testContext.setProgressState(ProgressState.STOPPED);
						throw e;
					} catch (CoreException e) {
						getLogger().trace("Unexpected", e);
						state.setStalled(false);
						testContext.setTestResult(Result.ERROR);
						state.addError(e);
					}

					testContext.setProgressState(ProgressState.COMPLETED);
				}

				if (getLogger().isDebugEnabled()) {
					final ITestElement element = elementRef[0];
					final String message = String.format("%s (%s;%s;%.3fs)",
							element, element.getTestResult(), element
									.getProgressState(true), element
									.getElapsedTimeInSeconds());
					getLogger().debug(message);
				}

				getLogger().debug("Run done");

				state.seekNext();

			} else {
				state.setUndoContext(null);

				if (state.hasError()) {
					final String message = "Failed to run test adapter(s), least one actor failed to run";
					state.reportError(message);
				}
			}

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}
}
