package de.fernunihagen.rtt.internal.runtime;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;

import de.fernunihagen.rtt.framework.ITestActor;
import de.fernunihagen.rtt.framework.ITestActorState;
import de.fernunihagen.rtt.framework.ITestConfiguration;
import de.fernunihagen.rtt.framework.ITestDescription;
import de.fernunihagen.rtt.framework.ITestElementContainer;
import de.fernunihagen.rtt.framework.TestActorState;

import de.fernunihagen.rtt.framework.model.TestContext;
import de.fernunihagen.rtt.internal.NotCancelableProgressMonitor;

/**
 * Test runner run state.
 * 
 * @param <A>
 *            Test actor type
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class TestRunnerRunState<A extends ITestActor> extends
		TestActorState {

	@Invar("null != $target")
	private final List<String> _loadFailures;
	@Invar("null != $target")
	protected final List<A> _actors;
	@Invar("null != $target")
	private final Map<A, ITestActorState> _states;

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				TestRunnerRunState.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Creates a test runner run state.
	 * 
	 * @param originator
	 *            Originator
	 * @param actors
	 *            Map of available actors
	 * @param actorConfigurations
	 *            List of actor configurations
	 */
	@Pre("null != originator && null != provider && null != actors")
	public TestRunnerRunState(final ITestActor originator,
			final Map<String, ? extends A> actors,
			final ITestConfiguration[] actorConfigurations) {
		super(originator);

		_loadFailures = new ArrayList<String>();
		_actors = new ArrayList<A>();
		_states = new HashMap<A, ITestActorState>();

		for (ITestConfiguration actorConfiguration : actorConfigurations) {
			final String name = actorConfiguration.get(
					ITestDescription.KEY_NAME, (String) null);

			Assert.isNotNull(name);

			final A actor = actors.get(name);
			if (null != actor) {
				_actors.add(actor);
			} else {
				_loadFailures.add(name);
			}
		}
	}

	@Override
	public void reset() {
		for (ITestActorState state : _states.values()) {
			state.reset();
		}
		_states.clear();
	}

	private void reportError(final CoreException[] exceptions,
			final String message) throws CoreException {
		getRTTCorePlugin().reportError(exceptions, message);
	}

	@Override
	public void initialize(final ITestElementContainer parent,
			final IProgressMonitor pm) throws CoreException {
		try {
			final SubMonitor monitor = SubMonitor.convert(pm,
					"Forcing initialization", _actors.size());
			final List<CoreException> errors = new ArrayList<CoreException>();

			for (A actor : _actors) {
				if (monitor.isCanceled()) {
					throw new OperationCanceledException();
				}

				try {
					createContext(actor, parent, monitor.newChild(1));
				} catch (CoreException e) {
					errors.add(e);
				}
			}

			if (!errors.isEmpty()) {
				reportError(errors.toArray(new CoreException[errors.size()]),
						"Error occured while creating test actor run state(s)");
			}
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	/**
	 * Creates a test context for given actor.
	 * 
	 * @param actor
	 *            Test actor
	 * @param parent
	 *            Parent test element container
	 * @param pm
	 *            The progress monitor to use for reporting progress to the
	 *            user. It is the caller's responsibility to call done() on the
	 *            given monitor. Accepts <code>null</code>, indicating that no
	 *            progress should be reported and that the operation cannot be
	 *            canceled
	 * 
	 * @return Test context
	 * 
	 * @throws CoreException
	 *             <ul>
	 *             <li>If an error occurs</li>
	 *             </ul>
	 */
	@Pre("null != actor && null != parent")
	@Post("null != $return")
	public final TestContext createContext(final A actor,
			final ITestElementContainer parent, final IProgressMonitor pm)
			throws CoreException {

		try {
			TestContext context;

			ITestActorState state = _states.get(actor);
			if (null == state) {
				final SubMonitor monitor = SubMonitor.convert(
						new NotCancelableProgressMonitor(pm),
						"Creating new actor state", 2);

				if (getLogger().isDebugEnabled()) {
					final String message = String.format(
							"Creating new actor %s state", actor.getName());
					getLogger().debug(message);
				}

				state = actor.getStateFactory().create(parent.getSession(),
						monitor.newChild(1));

				context = new TestContext(state, parent);
				context.activate(monitor.newChild(1));

				getLogger().debug("Creating done");

				_states.put(actor, state);
			} else {
				context = new TestContext(state, parent);
			}

			return context;

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	/**
	 * Retrieves the list of actor identifier failed to load.
	 * 
	 * @return List of actor identifiers failed to load
	 */
	@Post("null != $return")
	public final String[] getLoadFailures() {
		return _loadFailures.toArray(new String[_loadFailures.size()]);
	}

	/**
	 * Retrieves an empty actor array.
	 * 
	 * @return Empty actor array
	 */
	@Post("null != $return")
	protected abstract A[] getEmptyArray();

	/**
	 * Retrieves the list of actors loaded.
	 * 
	 * @return List of actors loaded
	 */
	@Post("null != $return")
	public final A[] getActors() {
		return _actors.toArray(getEmptyArray());
	}
}
