package de.fernunihagen.rtt.internal.runtime;

import static de.fernunihagen.rtt.framework.RTTCore.getRefactoringTestRunner;
import static de.fernunihagen.rtt.framework.RTTCore.getRepositoryProvider;

import static de.fernunihagen.rtt.internal.RTTCorePlugin.getRTTCorePlugin;
import static de.fernunihagen.rtt.internal.ResourceUtil.getResourceUtil;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRunnable;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.SubMonitor;

import de.fernunihagen.rtt.framework.ITestActorState;
import de.fernunihagen.rtt.framework.ITestConfiguration;
import de.fernunihagen.rtt.framework.ITestDescription;
import de.fernunihagen.rtt.framework.ITestElementContainer;
import de.fernunihagen.rtt.framework.XMLTestConfiguration;

import de.fernunihagen.rtt.framework.model.ResourceTestElement;
import de.fernunihagen.rtt.framework.model.TestContext;
import de.fernunihagen.rtt.framework.model.TestRootContext;

import de.fernunihagen.rtt.framework.repository.DiffTestElement;
import de.fernunihagen.rtt.framework.repository.IDiff;
import de.fernunihagen.rtt.framework.repository.IDiffVisitor;
import de.fernunihagen.rtt.framework.repository.IRepository;
import de.fernunihagen.rtt.framework.repository.IRepositoryProvider;

import de.fernunihagen.rtt.framework.testadapter.IRefactoringTestRunner;
import de.fernunihagen.rtt.framework.testadapter.RefactoringTestElement;

import de.fernunihagen.rtt.framework.util.IOverwriteQuery;
import de.fernunihagen.rtt.internal.NotCancelableProgressMonitor;

/**
 * The refactoring tool test application.
 * 
 * @author Osama El Hosami
 */
@Contract
public class ToolTestApplication extends AbstractTestApplication {

	private boolean _fDispose;
	private ITestConfiguration _configuration;
	@Invar("null != $return")
	private final Map<IProject, IRepository> _scope;

	private static final class LoggerHolder {
		private static final Logger _DEFAULT = getRTTCorePlugin().getLogger(
				ToolTestApplication.class.getName());
	}

	@Post("null != $return")
	private Logger getLogger() {
		return LoggerHolder._DEFAULT;
	}

	/**
	 * Creates a refactoring tool test application.
	 */
	public ToolTestApplication() {
		super();
		_configuration = null;
		_scope = new HashMap<IProject, IRepository>();
		_fDispose = false;
	}

	private IStatus createError(final String message, final Throwable e) {
		return getRTTCorePlugin().createError(message, e);
	}

	private void reportError(final CoreException[] exceptions,
			final String message) throws CoreException {
		getRTTCorePlugin().reportError(exceptions, message);
	}

	private void reportError(final String message, final Throwable exception)
			throws CoreException {
		getRTTCorePlugin().reportError(message, exception);
	}

	@Pre("null != filename")
	@Post("null != $return")
	private ITestConfiguration loadConfiguration(final String filename)
			throws CoreException {
		InputStreamReader reader = null;
		ITestConfiguration configuration = null;
		try {
			reader = new InputStreamReader(new BufferedInputStream(
					new FileInputStream(filename)), "UTF-8");

			configuration = XMLTestConfiguration.createReadRoot(reader);
		} catch (UnsupportedEncodingException e) {
			reportError("Unexpected file format", e);
		} catch (FileNotFoundException e) {
			reportError("File not found", e);
		} finally {
			getResourceUtil().close(reader);
		}

		return configuration;
	}

	@Override
	@Pre
	protected void parseArguments(final String[] args) throws CoreException {
		getLogger().debug(args);

		String testconfiguration = null;
		boolean fMatch = false;
		for (String arg : args) {
			if (fMatch) {
				testconfiguration = arg;
				break;
			}
			if (arg.equals("-testconfiguration")) {
				fMatch = true;
			}
		}

		if (null == testconfiguration) {
			reportError("Missing test configuration argument", null);
		}

		getLogger().debug(testconfiguration);
		_configuration = loadConfiguration(testconfiguration);
	}

	@Post("null != $return")
	private IWorkspace getWorkspace() {
		return getRTTCorePlugin().getWorkspace();
	}

	private void logWarning(final String message) {
		getRTTCorePlugin().logWarning(message);
	}

	@Override
	@Pre
	protected void importProjects(final IProgressMonitor pm)
			throws CoreException {
		getWorkspace().run(new IWorkspaceRunnable() {
			@Override
			public void run(final IProgressMonitor monitor)
					throws CoreException {
				try {
					importProjects(getRepositoryProvider(), getProjects(),
							monitor);
				} catch (CoreException e) {
					getLogger().trace("Unexpected", e);

					if (_scope.isEmpty()) {
						throw e;
					} else {
						logWarning("Continuing with incomplete test project set");
					}
				}
			}
		}, pm);
	}

	@Pre("null != projects && null != repository")
	private void addProjects(final IProject[] projects,
			final IRepository repository) {
		for (IProject project : projects) {
			if (getLogger().isDebugEnabled()) {
				final String message = String.format("Adding project %s of %s",
						project.getName(), repository.getName());
				getLogger().debug(message);
			}

			_scope.put(project, repository);
		}
	}

	@Pre("null != errors")
	private void validateProjects(final List<CoreException> errors) {
		IProject[] allProjects = _scope.keySet().toArray(
				new IProject[_scope.keySet().size()]);

		for (IProject project : allProjects) {
			if (!project.exists()) {
				_scope.remove(project);

				errors.add(new CoreException(createError("Failed to checkout "
						+ project.getName(), null)));
			}
		}
	}

	@Pre("null != repositoryProvider && null != configuration")
	private void importProjects(final IRepositoryProvider repositoryProvider,
			final ITestConfiguration configuration, final IProgressMonitor pm)
			throws CoreException {
		try {
			_scope.clear();

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Importing projects", SubMonitor.UNKNOWN);

			getLogger().debug("Importing projects");

			final List<CoreException> errors = new ArrayList<CoreException>();
			final Set<String> referenceStrings = new HashSet<String>();
			final ITestConfiguration[] providers = configuration
					.getChildren(ITestDescription.TYPE_PROJECTS_PROVIDER);

			for (ITestConfiguration provider : providers) {
				final ITestConfiguration[] projects = provider
						.getChildren(ITestDescription.TYPE_PROJECTS_PROVIDER_PROJECT);

				for (ITestConfiguration project : projects) {
					referenceStrings.add(project.get(ITestDescription.KEY_NAME,
							(String) null));
				}

				try {
					final String id = provider.get(ITestDescription.KEY_NAME,
							(String) null);

					final IRepository repository = repositoryProvider
							.getRepository(id);

					final String[] references = referenceStrings
							.toArray(new String[referenceStrings.size()]);

					final IProject[] allProjects = repository.checkout(
							references, new IOverwriteQuery() {
								@Override
								public StatusCode queryOverwrite(
										String pathString) {
									return StatusCode.ALL;
								}

							}, monitor.newChild(1));

					addProjects(allProjects, repository);

					referenceStrings.clear();
				} catch (CoreException e) {
					getLogger().trace("Unexpected", e);
					errors.add(e);
				}
			}

			validateProjects(errors);

			getLogger().debug("Importing done");

			if (!errors.isEmpty()) {
				if (errors.size() == 1) {
					throw errors.get(0);
				} else {
					reportError(
							errors.toArray(new CoreException[errors.size()]),
							"The following errors occurred while importing projects. Some projects may not be loaded.");
				}
			}
		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Post("null != $return")
	private ITestConfiguration getProjects() throws CoreException {
		ITestConfiguration configuration = getConfiguration().getChild(
				ITestDescription.TYPE_PROJECTS);
		if (null == configuration) {
			reportError("Configuration error, missing test project section",
					null);
		}
		return configuration;
	}

	@Pre("null != diffs")
	private void collectModifications(final List<IDiff> diffs,
			final IProgressMonitor pm) throws CoreException {
		try {
			diffs.clear();

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Collecting modifications", SubMonitor.UNKNOWN);

			final Set<Entry<IProject, IRepository>> entries = _scope.entrySet();
			final IDiffVisitor visitor = new IDiffVisitor() {
				@Override
				public boolean visit(final IDiff diff) {
					diffs.add(diff);
					monitor.worked(1);
					return !monitor.isCanceled();
				}
			};

			for (Entry<IProject, IRepository> entry : entries) {
				if (monitor.isCanceled()) {
					throw new OperationCanceledException();
				}

				monitor.subTask(entry.getKey().getName());

				final IRepository repository = entry.getValue();
				final IResource[] resources = new IResource[] { entry.getKey() };

				repository.refreshLocal(resources, IResource.DEPTH_INFINITE,
						monitor.newChild(1));

				repository.getSubscriber().accept(resources,
						IResource.DEPTH_INFINITE, visitor);
			}

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Pre("null != context && null != diffs")
	private void addModifications(final ITestElementContainer context,
			final List<IDiff> diffs) {

		final ITestElementContainer container = context.createChildContainer(
				diffs.size(), "Modifications");

		container.setTestResult(Result.NONE);
		container.setProgressState(ProgressState.RUNNING);

		for (IDiff diff : diffs) {
			final DiffTestElement element = new DiffTestElement(container, diff);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format("%s (%s;%s;%.3fs)",
						element, element.getTestResult(), element
								.getProgressState(true), element
								.getElapsedTimeInSeconds());
				getLogger().debug(message);
			}
		}

		container.setProgressState(ProgressState.COMPLETED);

		if (getLogger().isDebugEnabled()) {
			final String message = String.format("%s (%s;%s;%.3fs)", container,
					container.getTestResult(),
					container.getProgressState(true), container
							.getElapsedTimeInSeconds());
			getLogger().debug(message);
		}
	}

	@Pre("null != context && null != diffs")
	private void revertChanges(final ITestElementContainer context,
			final List<IDiff> diffs, final IProgressMonitor pm)
			throws CoreException {

		try {
			final Set<IProject> projects = new HashSet<IProject>();

			for (IDiff diff : diffs) {
				final IResource resource = getWorkspace().getRoot().findMember(
						diff.getPath());
				projects.add(resource.getProject());
			}

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Reverting changes", projects.size());

			final ITestElementContainer container = context
					.createChildContainer(projects.size(), "Rollback");

			container.setProgressState(ProgressState.RUNNING);

			for (IProject project : projects) {
				if (monitor.isCanceled()) {
					throw new OperationCanceledException();
				}

				monitor.subTask(project.getName());

				final ResourceTestElement element = new ResourceTestElement(
						container, project);

				element.setProgressState(ProgressState.RUNNING);

				_scope.get(project).revertChanges(new IResource[] { project },
						true, monitor.newChild(1));

				element.setTestResult(Result.OK);
				element.setProgressState(ProgressState.COMPLETED);

				if (getLogger().isDebugEnabled()) {
					final String message = String.format("%s (%s;%s;%.3fs)",
							element, element.getTestResult(), element
									.getProgressState(true), element
									.getElapsedTimeInSeconds());
					getLogger().debug(message);
				}
			}

			diffs.clear();
			container.setProgressState(ProgressState.COMPLETED);

			if (getLogger().isDebugEnabled()) {
				final String message = String.format("%s (%s;%s;%.3fs)",
						container, container.getTestResult(), container
								.getProgressState(true), container
								.getElapsedTimeInSeconds());
				getLogger().debug(message);
			}

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Override
	public void runImpl(final IProgressMonitor pm) throws CoreException {
		try {

			final SubMonitor monitor = SubMonitor.convert(pm,
					"Running test runner", SubMonitor.UNKNOWN);

			final List<IDiff> modifications = new ArrayList<IDiff>();
			final IRefactoringTestRunner runner = getRefactoringTestRunner();
			final ITestActorState state = runner.getStateFactory().create(this,
					new NotCancelableProgressMonitor(monitor.newChild(1)));

			boolean fCompleted;
			// Force pre-condition-check of first test run
			Result initialResult = Result.UNDEFINED;
			TestContext context = new TestContext(state, this);

			context.activate(monitor.newChild(1));

			do {
				monitor.setWorkRemaining(100);

				context = new TestRootContext(state, this, initialResult);

				fCompleted = context.run(monitor.newChild(1));

				if (context.getTarget() instanceof RefactoringTestElement) {
					collectModifications(modifications, monitor.newChild(1));
				}
				// else no refactoring has been executed

				if (!modifications.isEmpty()) {
					boolean fReverted = false;

					// Force pre-condition-check of next test run
					initialResult = Result.UNDEFINED;
					addModifications(context, modifications);

					try {
						if (runner.canUndoChange(context)) {
							runner.undoChange(context, monitor.newChild(1));

							collectModifications(modifications, null);
							fReverted = modifications.isEmpty();
						}
					} catch (CoreException e) {
						getLogger().trace("Unexpected", e);
					}

					if (!fReverted) {
						revertChanges(context, modifications, monitor
								.newChild(1));
					}
				} else {
					// Skip pre-condition-check of next test run - no
					// modifications
					initialResult = Result.NONE;
				}

			} while (!fCompleted);

		} finally {
			if (null != pm) {
				pm.done();
			}
		}
	}

	@Override
	protected boolean reset() {
		for (IProject project : getScope()) {
			try {
				project.open(null);
				IRepository repository = _scope.get(project);
				IResource[] resources = new IResource[] { project };

				repository.revertChanges(resources, true, null);
				repository.refreshLocal(resources, IResource.DEPTH_INFINITE,
						null);
			} catch (CoreException e) {
				setRequestProjects(true);
				break;
			}
		}

		return super.reset();
	}

	@Override
	@Post
	public ITestConfiguration getConfiguration() {
		return _configuration;
	}

	@Override
	@Post
	public IProject[] getScope() {
		return _scope.keySet().toArray(new IProject[_scope.keySet().size()]);
	}

	@Override
	@Post
	public String getLabel() {
		return "RTT Application";
	}

	@Override
	public void dispose() {
		setProgressState(ProgressState.STOPPED);
		_fDispose = true;
		setProgressState(ProgressState.NOT_STARTED);
	}

	@Override
	public boolean isDisposed() {
		return _fDispose;
	}
}
