package de.fernunihagen.rtt.internal.util;

import java.lang.reflect.Method;

import java.util.Arrays;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Pre;

/**
 * Matches a public method by exact name and/ or a set of signature attributes.
 * In case any attribute is set to null, it is assumed as optional and left out
 * in comparison - the first best matching method is returned.
 * 
 * @author Osama El Hosami
 */
@Contract
public final class BestPublicMethodMatcher implements IMethodMatcher {

	private static final class BestMethodMatcherHolder {
		private static final BestPublicMethodMatcher _DEFAULT = new BestPublicMethodMatcher();
	}

	private BestPublicMethodMatcher() {
	}

	/**
	 * Retrieves the shared instance.
	 * 
	 * @return The shared instance
	 */
	public static IMethodMatcher getBestPublicMethodMatcher() {
		return BestMethodMatcherHolder._DEFAULT;
	}

	/**
	 * Test is method matches given name and signature attributes.
	 * 
	 * @param method
	 *            Method
	 * @param paramTypes
	 *            Parameter argument types
	 * @param returnType
	 *            Return type
	 * @param methodName
	 *            Method name
	 * 
	 * @return True if matches otherwise false
	 */
	@Pre("null != method")
	public boolean matches(final Method method, final Class<?>[] paramTypes,
			final Class<?> returnType, final String methodName) {

		return (null == returnType || method.getReturnType().equals(returnType))
				&& (null == paramTypes || Arrays.equals(method
						.getParameterTypes(), paramTypes))
				&& (null == methodName || method.getName()
						.compareTo(methodName) == 0);
	}

	@Override
	@Pre()
	public Method lookup(final Class<?> clazz, final Class<?>[] paramTypes,
			final Class<?> returnType, final String methodName) {
		Method matching = null;

		if (null != methodName && null != paramTypes && null != returnType) {
			try {
				matching = clazz.getMethod(methodName, paramTypes);
			} catch (SecurityException e) {
			} catch (NoSuchMethodException e) {
			}
		} else {
			for (final Method method : clazz.getMethods()) {
				if (matches(method, paramTypes, returnType, methodName)) {
					matching = method;
					break;
				}
			}
		}

		return matching;
	}

}
