package de.fernunihagen.rtt.internal.util;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

/**
 * Base class of conditions, used in conjunction with a function objects
 * (functor).
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class Condition {

	/**
	 * Condition representing true.
	 */
	public static final Condition TRUE = new ConditionTrue();

	/**
	 * Condition representing false.
	 */
	public static final Condition FALSE = new ConditionFalse();

	/**
	 * Interprets given boolean value and returns associated condition.
	 * 
	 * @param fValue
	 *            Condition value
	 * 
	 * @return Condition
	 */
	public static Condition test(boolean fValue) {
		return (fValue) ? Condition.TRUE : Condition.FALSE;
	}

	/**
	 * Returns the negated condition.
	 * 
	 * @return Negated condition
	 */
	@Post("null != $return")
	public abstract Condition not();

	/**
	 * Returns a condition as result of an and operation of this with another.
	 * condition
	 * 
	 * @return Resulting condition
	 */
	@Pre("null != other")
	@Post("null != $return")
	public abstract Condition and(Condition other);

	/**
	 * Returns a condition as result of an or operation of this with another.
	 * condition
	 * 
	 * @return Resulting condition
	 */
	@Pre("null != other")
	@Post("null != $return")
	public abstract Condition or(Condition other);

	/**
	 * Returns a condition as result of a xor operation of this with another.
	 * condition
	 * 
	 * @return Resulting condition
	 */
	@Pre("null != other")
	@Post("null != $return")
	public abstract Condition xor(Condition other);

	/**
	 * Invokes a functor depending on evaluation of condition.
	 * 
	 * @param <R>
	 *            Result type
	 * @param <X>
	 *            Exception type
	 * @param functor
	 *            Function object
	 * @param args
	 *            List of arguments
	 * 
	 * @return Result
	 */
	@Pre("null != functor")
	public abstract <R, X extends Throwable> R invoke(
			IConditionalCommand<R, X> functor, Object... args) throws X;
}
