package de.fernunihagen.rtt.internal.util;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Pre;

/**
 * The method delegate proxy (for duck-typing), used as dynamic invocation
 * handler - described in the article "A Java Programmer Looks at C# Delegates"
 * by <code>Steven M. Lewis</code> and <code>Wilhelm Fitzpatrick</code>.
 * 
 * @param <T>
 *            Delegate interface
 * @param <R>
 *            Result type
 * @param <X>
 *            Exception type
 * 
 * @author Osama El Hosami
 */
@Contract
public final class DelegateProxy<T, R, X extends Throwable> implements
		IDelegate<R, X>, InvocationHandler {
	@Invar("null != $target")
	private final Delegator<T> _delegator;
	@Invar("null != $target")
	private final Class<?> _targetClass;
	@Invar("null != $target")
	private final Class<R> _returnType;
	@Invar("null != $target")
	private final Method _method;
	private final Object _target;

	/**
	 * Creates a delegate proxy.
	 * 
	 * @param target
	 *            Target with a bindable method
	 * @param paramTypes
	 *            Delegate parameter types
	 * @param returnType
	 *            Method return type
	 * @param methodName
	 *            Name of the method or null
	 * @param delegator
	 *            Delegate factory
	 * 
	 * @throws IllegalAccessException
	 *             If method not found
	 */
	@Pre("null != returnType && null != delegator")
	public DelegateProxy(final Object target, final Class<?>[] paramTypes,
			final Class<R> returnType, final String methodName,
			final Delegator<T> delegator) throws IllegalAccessException {

		_returnType = returnType;
		_delegator = delegator;
		_targetClass = delegator.getTargetInterface(target);
		_target = delegator.isClassInstance(target) ? null : target;
		_method = delegator.lookup(_targetClass, paramTypes, _returnType,
				methodName);

		if (_method == null) {
			throw new IllegalAccessException();
		}
	}

	/**
	 * The interface of all delegates created by delegator.
	 * 
	 * @return Delegate interface or null
	 */
	public Class<T> getInterface() {
		return _delegator.getInterface();
	}

	@Override
	public R invoke(final Object... args) throws X {
		R result = null;
		try {
			result = _returnType.cast(_method.invoke(_target, args));
		} catch (IllegalAccessException e) {
			throw new RuntimeException(e);
		} catch (InvocationTargetException e) {
			throw new RuntimeException(e);
		}
		return result;
	}

	@Override
	public Object invoke(final Object proxy, final Method method,
			final Object[] args) throws Throwable {
		return invoke(args);
	}
}
