package de.fernunihagen.rtt.internal.util;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.Proxy;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

import static de.fernunihagen.rtt.internal.util.BestPublicMethodMatcher.getBestPublicMethodMatcher;

/**
 * Factory used for delegate creation - described in the article "A Java
 * Programmer Looks at C# Delegates" by <code>Steven M. Lewis</code> and
 * <code>Wilhelm Fitzpatrick</code>.
 * 
 * @param <T>
 *            Delegate interface
 * 
 * @author Osama El Hosami
 */
@Contract
public final class Delegator<T> implements IMethodMatcher {
	private final Class<T> _itf;
	private final Class<?>[] _paramTypes;
	@Invar("null != $target")
	private final IMethodMatcher _methodMatcher;

	@Pre("null != methodMatcher")
	private Delegator(final Class<T> itf, final Class<?>[] paramTypes,
			IMethodMatcher methodMatcher) {

		if (!isInterfaceType(itf)) {
			throw new IllegalArgumentException("itf not a interface type");
		}

		if (methodMatcher instanceof Delegator<?>) {
			methodMatcher = Delegator.class.cast(methodMatcher).getMatcher();
		}

		_itf = itf;
		_paramTypes = paramTypes;
		_methodMatcher = methodMatcher;
	}

	/**
	 * Test if class type is a interface or null
	 * 
	 * @param <T>
	 *            Type
	 * 
	 * @param itf
	 *            Class type
	 * 
	 * @return True if satisfied otherwise false
	 */
	public static <T> boolean isInterfaceType(final Class<T> itf) {
		return itf == null || Modifier.isInterface(itf.getModifiers());
	}

	/**
	 * Creates a delegator, which can be used for creation of delegates of given
	 * interface type.
	 * 
	 * @param <T>
	 *            Delegate interface
	 * @param itf
	 *            Target object interface of delegate or null
	 * @param methodMatcher
	 *            Matcher used to lookup method by name
	 * 
	 * @throws IllegalArgumentException
	 *             <ul>
	 *             <li>
	 *             <li>If itf is non-null and not a interface type</li>
	 *             </ul>
	 */
	public static <T> Delegator<T> getDelegator(final Class<T> itf,
			final IMethodMatcher methodMatcher) {
		return new Delegator<T>(itf, null, methodMatcher);
	}

	/**
	 * Creates a delegator, which can be used for creation of delegates of given
	 * prototype.
	 * 
	 * @param <T>
	 *            Delegate interface
	 * @param paramTypes
	 *            Delegate parameter types
	 * @param methodMatcher
	 *            Matcher used to lookup method by name
	 */
	public static <T> Delegator<T> getDelegator(final Class<?>[] paramTypes,
			final IMethodMatcher methodMatcher) {
		return new Delegator<T>(null, paramTypes, methodMatcher);
	}

	/**
	 * Creates a delegator, which can be used for creation of delegates of given
	 * prototype.
	 * 
	 * @param <T>
	 *            Delegate interface
	 * @param paramTypes
	 *            Delegate parameter types
	 */
	public static <T> Delegator<T> getDelegator(final Class<?>[] paramTypes) {
		return getDelegator(paramTypes, getBestPublicMethodMatcher());
	}

	/**
	 * Creates a delegator, which can be used for creation of delegates of given
	 * interface type.
	 * 
	 * @param <T>
	 *            Delegate interface
	 * @param itf
	 *            Target object interface of delegate or null
	 * 
	 * @throws IllegalArgumentException
	 *             <ul>
	 *             <li>
	 *             <li>If itf is non-null and not a interface type</li>
	 *             </ul>
	 */
	public static <T> Delegator<T> getDelegator(final Class<T> itf) {
		return getDelegator(itf, getBestPublicMethodMatcher());
	}

	/**
	 * Creates a delegator, which can be used for delegate creation.
	 * 
	 * @param methodMatcher
	 *            Matcher used to lookup method by name
	 */
	public static <T> Delegator<T> getDelegator(
			final IMethodMatcher methodMatcher) {
		return new Delegator<T>(null, null, methodMatcher);
	}

	/**
	 * Creates a delegator, which can be used for delegate creation.
	 */
	public static <T> Delegator<T> getDelegator() {
		return getDelegator(getBestPublicMethodMatcher());
	}

	/**
	 * Retrieves the method matcher used to find a method by name.
	 * 
	 * @return Matcher instance
	 */
	@Post("null != $return")
	public IMethodMatcher getMatcher() {
		return _methodMatcher;
	}

	/**
	 * The interface of all delegates created by delegator.
	 * 
	 * @return Delegate interface or null
	 */
	@Post("$this.isInterfaceType($return)")
	public Class<T> getInterface() {
		return _itf;
	}

	@Override
	@Pre()
	public Method lookup(final Class<?> clazz, final Class<?>[] paramTypes,
			final Class<?> returnType, final String methodName) {
		return _methodMatcher.lookup(clazz, paramTypes, returnType, methodName);
	}

	/**
	 * Test if target is a class instance - in this case delegate requested to
	 * bind on target, are assumed to have the owner scope classifier.
	 * 
	 * @param target
	 *            Target instance
	 * 
	 * @return True if satisfied otherwise false
	 */
	public boolean isClassInstance(final Object target) {
		return target != null && target instanceof Class<?>;
	}

	/**
	 * Retrieves the interface of a target instance, that can be used to bind
	 * method.
	 * 
	 * @param target
	 *            Target instance
	 * 
	 * @return Interface of target instance
	 */
	@Pre("null != target")
	@Post("null != $return")
	public Class<?> getTargetInterface(final Object target) {
		final Class<?> targetClass;

		if (isClassInstance(target)) {
			targetClass = (Class<?>) target;
		} else {
			targetClass = target.getClass();
		}

		return targetClass;
	}

	/**
	 * Creates a dynamic delegate proxy.
	 * 
	 * @param <R>
	 *            Return type
	 * @param <X>
	 *            Exception type
	 * @param target
	 *            Target object instance or null
	 * @param itf
	 *            Delegate interface
	 * @param proxy
	 *            Static delegate proxy
	 * 
	 * @return The delegate, wrapped by a dynamic proxy
	 */
	@SuppressWarnings("unchecked")
	@Pre("null != itf && null != proxy && $this.isInterfaceType(itf)")
	@Post("null != $return")
	protected <R, X extends Throwable> IDelegate<R, X> createDynamicProxy(
			final Object target, final Class<T> itf,
			final DelegateProxy<T, R, X> proxy) {

		final Class<?>[] interfaces = { itf, IDelegate.class };

		return (IDelegate) Proxy.newProxyInstance(getTargetInterface(target)
				.getClassLoader(), interfaces, proxy);
	}

	/**
	 * Binds a delegate on target object method by name.
	 * 
	 * @param <R>
	 *            Return type
	 * @param <X>
	 *            Exception type
	 * @param target
	 *            Target with a bindable method
	 * @param methodName
	 *            Name of the method
	 * @param returnType
	 *            Method return type
	 * 
	 * @return The delegate. If getInterface() is non-null the returned Delegate
	 *         will be a dynamic proxy implementing that interface
	 * 
	 * @throws IllegalAccessException
	 *             <ul>
	 *             <li>
	 *             If method not found</li>
	 *             <li>If method is inaccessible</li>
	 *             </ul>
	 */
	@Pre("null != target && null != returnType")
	@Post("null != $return")
	public <R, X extends Throwable> IDelegate<R, X> bind(final Object target,
			final String methodName, final Class<R> returnType)
			throws IllegalAccessException {
		return bind(getInterface(), returnType, target, methodName);
	}

	/**
	 * Faces a bindable method of target object by name, so that it supports
	 * desired delegate interface.
	 * 
	 * @param <R>
	 *            Return type
	 * @param target
	 *            Target with a bindable method
	 * @param methodName
	 *            Name of the method
	 * @param returnType
	 *            Method return type
	 * 
	 * @return The delegate
	 * 
	 * @throws UnsupportedOperationException
	 *             <ul>
	 *             <li>
	 *             If no interface is defined</li>
	 *             </ul>
	 * @throws IllegalAccessException
	 *             <ul>
	 *             <li>
	 *             If method not found</li>
	 *             <li>If method is inaccessible</li>
	 *             </ul>
	 */
	@Pre("null != target && null != returnType")
	@Post("null != $return")
	public <R> T face(final Object target, final String methodName,
			final Class<R> returnType) throws IllegalAccessException {
		final Class<T> itf = getInterface();
		if (null == itf) {
			throw new UnsupportedOperationException("no interface is defined");
		}

		return itf.cast(bind(itf, returnType, target, methodName));
	}

	@Pre("null != target && null != returnType")
	@Post("null != $return")
	private final <R, X extends Throwable> IDelegate<R, X> bind(
			final Class<T> itf, final Class<R> returnType, final Object target,
			final String methodName) throws IllegalAccessException {

		final DelegateProxy<T, R, X> delegate = new DelegateProxy<T, R, X>(
				target, _paramTypes, returnType, methodName, this);

		if (itf != null && !itf.equals(IDelegate.class)) {
			return createDynamicProxy(target, itf, delegate);
		}

		return delegate;
	}
}
