package de.fernunihagen.rtt.internal.util;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Post;
import org.contract4j5.contract.Pre;

/**
 * Last recently used cache - implementation is not thread-safe.
 * 
 * @author Osama El Hosami
 * 
 * @param <K>
 *            Key type
 * @param <V>
 *            Value type
 */
@Contract
public final class LRUCache<K, V> extends AbstractCache<K, V> {
	private static final long serialVersionUID = -8378162013731833095L;

	@Invar("0 < $target")
	private final int _capacity;
	@Invar("null != $target")
	private final LinkedHashMap<K, V> _map;

	/**
	 * Creates a LRU cache.
	 * 
	 * @param capacity
	 *            Cache capacity
	 */
	@Pre("0 < capacity")
	public LRUCache(final int capacity) {
		this(capacity, new NullFactory<V, ICache<K, V>>());
	}

	/**
	 * Creates a LRU cache.
	 * 
	 * @param capacity
	 *            Cache capacity
	 * @param provider
	 *            Cache value provider
	 */
	@Pre("0 < capacity && null != provider")
	public LRUCache(final int capacity, final IFactory<V, ICache<K, V>> provider) {
		_capacity = capacity;
		_map = new LinkedHashMap<K, V>(capacity + 1, 0.75f, true) {
			private static final long serialVersionUID = 1526375191423975664L;

			@Override
			protected final boolean removeEldestEntry(
					final Map.Entry<K, V> eldest) {
				return super.size() > capacity;
			}
		};
		setProvider(provider);
	}

	@Post
	@Override
	public int getCapacity() {
		return _capacity;
	}

	@Override
	public void clear() {
		_map.clear();
	}

	@Override
	public boolean containsKey(final Object arg0) {
		return _map.containsKey(arg0);
	}

	@Override
	public boolean containsValue(final Object arg0) {
		return _map.containsValue(arg0);
	}

	@Override
	public Set<java.util.Map.Entry<K, V>> entrySet() {
		return _map.entrySet();
	}

	@Override
	public V get(final Object key) {
		V value = _map.get(key);
		if (value == null && !containsKey(key)) {
			value = getProvider().create(this, key);
		}
		return value;
	}

	@Override
	public boolean isEmpty() {
		return _map.isEmpty();
	}

	@Override
	public Set<K> keySet() {
		return _map.keySet();
	}

	@Override
	public V put(final K arg0, final V arg1) {
		return _map.put(arg0, arg1);
	}

	@Override
	public void putAll(final Map<? extends K, ? extends V> arg0) {
		_map.putAll(arg0);
	}

	@Override
	public V remove(final Object arg0) {
		return _map.remove(arg0);
	}

	@Override
	public int size() {
		return _map.size();
	}

	@Override
	public Collection<V> values() {
		return _map.values();
	}
}
