package de.fernunihagen.rtt.internal.util;

import static de.fernunihagen.rtt.internal.util.ClassUtil.getClassUtil;

import de.fernunihagen.rtt.framework.util.IPropertyContainer;
import de.fernunihagen.rtt.framework.util.IPropertyObject;

/**
 * Property utility class.
 * 
 * @author Osama El Hosami
 */
public final class PropertyUtil {
	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> String getConvertedProperty(
			final IPropertyContainer<S> container, final String key,
			String defaultValue) {
		if (null == defaultValue) {
			defaultValue = "";
		}
		return container.getProperty(key, defaultValue).toString();
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> String getProperty(final IPropertyContainer<S> container,
			final String key, final String defaultValue) {
		return getClassUtil().as(container.getProperty(key, defaultValue),
				String.class);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> boolean getConvertedProperty(
			final IPropertyContainer<S> container, final String key,
			final boolean defaultValue) {
		return Boolean.parseBoolean(getConvertedProperty(container, key,
				Boolean.valueOf(defaultValue).toString()));
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> boolean getProperty(
			final IPropertyContainer<S> container, final String key,
			final boolean defaultValue) {
		return getClassUtil().as(container.getProperty(key, defaultValue),
				Boolean.class);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> int getConvertedProperty(
			final IPropertyContainer<S> container, final String key,
			final int defaultValue) {
		int value = defaultValue;

		try {
			value = Integer.parseInt(getConvertedProperty(container, key,
					Integer.valueOf(defaultValue).toString()));
		} catch (NumberFormatException e) {

		}

		return value;
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> int getProperty(final IPropertyContainer<S> container,
			final String key, final int defaultValue) {
		return getClassUtil().as(container.getProperty(key, defaultValue),
				Integer.class);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> short getConvertedProperty(
			final IPropertyContainer<S> container, final String key,
			final short defaultValue) {
		short value = defaultValue;

		try {
			value = Short.parseShort(getConvertedProperty(container, key, Short
					.valueOf(defaultValue).toString()));
		} catch (NumberFormatException e) {

		}

		return value;
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> short getProperty(final IPropertyContainer<S> container,
			final String key, final short defaultValue) {
		final Object value = container.getProperty(key, defaultValue);
		return getClassUtil().as(value, Short.class);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> byte getConvertedProperty(
			final IPropertyContainer<S> container, final String key,
			final byte defaultValue) {
		byte value = defaultValue;

		try {
			value = Byte.parseByte(getConvertedProperty(container, key, Byte
					.valueOf(defaultValue).toString()));
		} catch (NumberFormatException e) {

		}

		return value;
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> byte getProperty(final IPropertyContainer<S> container,
			final String key, final byte defaultValue) {
		final Object value = container.getProperty(key, defaultValue);
		return getClassUtil().as(value, Byte.class);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> double getConvertedProperty(
			final IPropertyContainer<S> container, final String key,
			final double defaultValue) {
		double value = defaultValue;

		try {
			value = Double.parseDouble(getConvertedProperty(container, key,
					Double.valueOf(defaultValue).toString()));
		} catch (NumberFormatException e) {

		}

		return value;
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> double getProperty(final IPropertyContainer<S> container,
			final String key, final double defaultValue) {
		final Object value = container.getProperty(key, defaultValue);
		return getClassUtil().as(value, Double.class);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> float getConvertedProperty(
			final IPropertyContainer<S> container, final String key,
			final float defaultValue) {
		float value = defaultValue;

		try {
			value = Float.parseFloat(getConvertedProperty(container, key, Float
					.valueOf(defaultValue).toString()));
		} catch (NumberFormatException e) {

		}

		return value;
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> float getProperty(final IPropertyContainer<S> container,
			final String key, final float defaultValue) {
		final Object value = container.getProperty(key, defaultValue);
		return getClassUtil().as(value, Float.class);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> long getConvertedProperty(
			final IPropertyContainer<S> container, final String key,
			final long defaultValue) {
		long value = defaultValue;

		try {
			value = Long.parseLong(getConvertedProperty(container, key, Long
					.valueOf(defaultValue).toString()));
		} catch (NumberFormatException e) {

		}

		return value;
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property container
	 * 
	 * @param container
	 *            Property container
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> long getProperty(final IPropertyContainer<S> container,
			final String key, final long defaultValue) {
		final Object value = container.getProperty(key, defaultValue);
		return getClassUtil().as(value, Long.class);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> String getConvertedProperty(final IPropertyObject<S> obj,
			final String key, String defaultValue) {
		return getConvertedProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> String getProperty(final IPropertyObject<S> obj,
			final String key, final String defaultValue) {
		return getProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> boolean getConvertedProperty(
			final IPropertyObject<S> obj, final String key, boolean defaultValue) {
		return getConvertedProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> boolean getProperty(final IPropertyObject<S> obj,
			final String key, final boolean defaultValue) {
		return getProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> int getConvertedProperty(final IPropertyObject<S> obj,
			final String key, final int defaultValue) {
		return getConvertedProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> int getProperty(final IPropertyObject<S> obj,
			final String key, final int defaultValue) {
		return getProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> short getConvertedProperty(final IPropertyObject<S> obj,
			final String key, final short defaultValue) {
		return getConvertedProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> short getProperty(final IPropertyObject<S> obj,
			final String key, final short defaultValue) {
		return getProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> byte getConvertedProperty(final IPropertyObject<S> obj,
			final String key, final byte defaultValue) {
		return getConvertedProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> byte getProperty(final IPropertyObject<S> obj,
			final String key, final byte defaultValue) {
		return getProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> double getConvertedProperty(final IPropertyObject<S> obj,
			final String key, final double defaultValue) {
		return getConvertedProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> double getProperty(final IPropertyObject<S> obj,
			final String key, final double defaultValue) {
		return getProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> float getConvertedProperty(final IPropertyObject<S> obj,
			final String key, final float defaultValue) {
		return getConvertedProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> float getProperty(final IPropertyObject<S> obj,
			final String key, final float defaultValue) {
		return getProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key converted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> long getConvertedProperty(final IPropertyObject<S> obj,
			final String key, final long defaultValue) {
		return getConvertedProperty(obj.getProperties(), key, defaultValue);
	}

	/**
	 * Returns the property value associated with the given key casted to the
	 * type of default value.
	 * 
	 * @param <S>
	 *            Subject type of property object
	 * 
	 * @param obj
	 *            Property object
	 * 
	 * @param key
	 *            Property key
	 * 
	 * @param defaultValue
	 *            Default value that will be returned if no property value was
	 *            associated with the given key
	 * 
	 * @return Property value associated with the given key
	 */
	public static <S> long getProperty(final IPropertyObject<S> obj,
			final String key, final long defaultValue) {
		return getProperty(obj.getProperties(), key, defaultValue);
	}
}
