package de.fernunihagen.rtt.internal.util;

import java.lang.ref.WeakReference;

import java.util.Set;

import java.util.concurrent.CopyOnWriteArraySet;
import java.util.concurrent.atomic.AtomicBoolean;

import org.contract4j5.contract.Contract;
import org.contract4j5.contract.Invar;
import org.contract4j5.contract.Pre;

import de.fernunihagen.rtt.framework.util.IObservable;
import de.fernunihagen.rtt.framework.util.IObserver;

/**
 * A generic observable object (subject). Implementation uses weak references to
 * manage registered observers - the weak references ensures that observers can
 * be garbaged, while they remain registered and considered to be unreachable
 * (or only "weakly reachable").
 * 
 * @param <S>
 *            Subject interface
 * @param <O>
 *            Observer interface
 * @param <A>
 *            Argument interface
 * 
 * @author Osama El Hosami
 */
@Contract
public abstract class Subject<S extends IObservable<S, O, A>, O extends IObserver<S, O, A>, A>
		implements IObservable<S, O, A> {

	@Contract
	private final class ObserverReference extends WeakReference<O> {
		@Pre("null != o")
		public ObserverReference(final O o) {
			super(o);
		}

		/**
		 * Test if observer reference is unreachable
		 * 
		 * @return true if satisfied otherwise false
		 */
		public boolean isUnreachable() {
			return null == get();
		}

		@Override
		public boolean equals(final Object o) {
			if (this == o) {
				return true;
			}
			if (null == o || !ObserverReference.class.isInstance(o)) {
				return false;
			}

			return o.equals(this);
		}

		@Pre("null != o")
		private boolean equals(final ObserverReference o) {
			final O observer = get();
			return null != o && observer.equals(o.get());
		}

		@Override
		public int hashCode() {
			final O observer = get();
			return (null != observer) ? observer.hashCode() : super.hashCode();
		}
	}

	@Invar("null != $target")
	private final Set<ObserverReference> _observers;
	@Invar("null != $target")
	private final AtomicBoolean _change;

	/**
	 * Creates the subject
	 */
	public Subject() {
		_observers = new CopyOnWriteArraySet<ObserverReference>();
		_change = new AtomicBoolean(false);
	}

	/**
	 * Marks this Observable object as having been changed; the hasChanged
	 * method will now return true.
	 */
	public void setChanged() {
		_change.set(true);
	}

	/**
	 * Indicates that this object has no longer changed, or that it has already
	 * notified all of its observers of its most recent change, so that the
	 * hasChanged method will now return false.
	 */
	public void clearChanged() {
		_change.set(false);
	}

	/**
	 * Retrieves the more precise this type
	 * 
	 * @return The more precise this type
	 */
	protected abstract S getThis();

	/**
	 * Tests weather an observer is already contained or not
	 * 
	 * @param o
	 *            An observer
	 * 
	 * @return True if contained otherwise false
	 */
	@Pre("null != o")
	public final boolean contains(final O o) {
		return _observers.contains(new ObserverReference(o));
	}

	@Override
	@Pre
	public final void addObserver(final O o) {
		_observers.add(new ObserverReference(o));
	}

	@Override
	@Pre
	public final int countObservers() {
		return _observers.size();
	}

	@Override
	public final void deleteObserver(final O o) {
		final ObserverReference refObserver = new ObserverReference(o);

		for (ObserverReference ref : _observers) {
			if (ref.isUnreachable() || ref.equals(refObserver)) {
				_observers.remove(ref);
			}
		}
	}

	@Override
	@Pre
	public final void deleteObservers() {
		_observers.clear();
	}

	@Override
	@Pre
	public final void notifyObservers() {
		notifyObservers(null);
	}

	@Override
	@Pre
	public final void notifyObservers(final A arg) {
		if (hasChanged()) {
			final S s = getThis();
			for (ObserverReference ref : _observers) {
				final O observer = ref.get();
				if (!ref.isUnreachable()) {
					observer.update(s, arg);
				} else {
					_observers.remove(ref);
				}
			}
			clearChanged();
		}
	}

	@Override
	public final boolean hasChanged() {
		return _change.get();
	}
}
